/*
 * AFAPL2Compiler.java
 *
 * Created on 14 June 2006, 13:37
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package com.agentfactory.afapl2;

import com.agentfactory.afapl2.compiler.AFAPL2SyntaxError;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.agentfactory.afapl2.compiler.Phase1Visitor;
import com.agentfactory.afapl2.compiler.Phase2Visitor;
import com.agentfactory.afapl2.compiler.Phase3Visitor;
import com.agentfactory.afapl2.compiler.Phase4Visitor;
import com.agentfactory.afapl2.compiler.parser.LexicalAnalyser;
import com.agentfactory.afapl2.compiler.parser.ParseTreeGenerator;
import com.agentfactory.afapl2.compiler.parser.node.Node;
import com.agentfactory.afapl2.compiler.parser.node.ProgramNode;
import com.agentfactory.afapl2.compiler.parser.reader.FileReader;
import com.agentfactory.afapl2.compiler.parser.reader.URLReader;

/**
 *
 * @author Administrator
 */
public class AFAPL2Compiler {

    private List<AFAPL2SyntaxError> errors, warnings;
    private List<File> files;
    private List<String> targets;
    private ParseTreeGenerator generator;

    public class AFAPL2FilenameFilter implements java.io.FilenameFilter {

        public boolean accept( File file, String text ) {
            return text.endsWith( ".afapl2" );
        }
    }

    /** Creates a new instance of AFAPL2Compiler */
    public AFAPL2Compiler() {
        targets = new ArrayList<String>();
        files = new ArrayList<File>();
        errors = new ArrayList<AFAPL2SyntaxError>();
        warnings = new ArrayList<AFAPL2SyntaxError>();
        generator = new ParseTreeGenerator();
    }

    public void addTarget( String target ) {
        targets.add( target );
    }

    public boolean compileFiles() {
        // Start by expanding the targets...
        for ( String s : targets ) {
            expandTarget( s );
        }

        for ( File f : files ) {
            compileFile( f );
        }

        Iterator it = errors.iterator();
        while ( it.hasNext() ) {
            System.out.println( it.next() );
        }

        it = warnings.iterator();
        while ( it.hasNext() ) {
            System.out.println( it.next() );
        }

        System.out.println( "Errors=" + errors.size() + ", Warnings=" + warnings.size() );
        if ( errors.isEmpty() ) {
            System.out.println( "COMPILATION SUCCEEDED" );
        }
        else {
            System.out.println( "COMPILATION FAILED" );
        }
        return errors.isEmpty();
    }

    private void expandTarget( String target ) {
        File file = new File( target );
        if ( file.isFile() ) {
            if ( file.getName().endsWith( ".afapl2" ) ) {
                files.add( file );
            }
            else {
                System.out.println( "File will not be compiled: Invalid extension: " + file );
            }
        }
        if ( file.isDirectory() ) {
            File[] files = file.listFiles( new AFAPL2FilenameFilter() );
            for ( int i = 0; i < files.length; i++ ) {
                this.files.add( files[i] );
            }
        }
    }

    private void compileFile( File file ) {
        LexicalAnalyser analyser = new LexicalAnalyser( new FileReader( file ) );
        if ( generator.generateParseTree( analyser ) ) {
            ProgramNode node = generator.getRoot();

            // Prepare visitors
            Phase1Visitor p1Visitor = new Phase1Visitor( errors, warnings );
            Phase2Visitor p2Visitor = new Phase2Visitor( errors, warnings );
            Phase3Visitor p3Visitor = new Phase3Visitor( errors, warnings, p2Visitor );

            // Pass 1: IMPORT Statement Expansion
            node.accept( p1Visitor );

            // If errors were detected in the first pass, stop compiling
            if ( errors.size() > 0 ) {
                return;
            }

            // Expand IMPORTS (returns a list of secondary source files that are
            // associated with the primary source file via the import statement)
            List<Node> sources = expandImports( p1Visitor );
            if ( sources != null ) {
                // OKAY: We have some valid sources (a primary + imports)
                // Add the primary to the sources list.  From this point on
                // Any visitors will visit all source nodes...
                sources.add( 0, node );

                // Pass 2: Class Check + Activity List Generation + Ontology
                //         Generation
                ProgramNode pNode = null;
                Iterator it = sources.iterator();
                while ( it.hasNext() ) {
                    pNode = (ProgramNode) it.next();
                    pNode.accept( p2Visitor );
                }

                // If errors were detected in the second pass, stop compiling
                if ( errors.size() > 0 ) {
                    return;
                }

                // Pass 3: Activity & Ontology Checks
                it = sources.iterator();
                while ( it.hasNext() ) {
                    pNode = (ProgramNode) it.next();
                    pNode.accept( p3Visitor );
                }

                // If errors were detected in the third pass, stop compiling
                if ( errors.size() > 0 ) {
                    return;
                }

                // Pass 4: Create the deployment file
                String source = node.getSource();
                String target = source.substring( 0, source.lastIndexOf( "." ) ) + ".agent";

                PrintWriter out = null;
                try {
                    out = new PrintWriter( new FileWriter( new File( target ) ) );
                }
                catch ( IOException ioe ) {
                    System.out.println( "ERROR: Could not create file: " + target );
                    return;
                }
                Phase4Visitor p4Visitor = new Phase4Visitor( errors, warnings, out, p2Visitor );

                it = sources.iterator();
                while ( it.hasNext() ) {
                    pNode = (ProgramNode) it.next();
                    pNode.accept( p4Visitor );
                }
                out.close();
            }
        }
        else {
            errors.addAll( generator.getErrors() );
        }
    }

    private List<Node> expandImports( Phase1Visitor iVisitor ) {
        List<String> imports = iVisitor.getImports();
        List<Node> sources = new ArrayList<Node>();
        for ( int i = 0; i < imports.size(); i++ ) {
            LexicalAnalyser analyser = new LexicalAnalyser( new URLReader( imports.get( i ) ) );
            if ( generator.generateParseTree( analyser ) ) {
                ProgramNode node = generator.getRoot();
                node.accept( iVisitor );
                sources.add( node );
            }
            else {
                errors.addAll( generator.getErrors() );
                return null;
            }
        }

        return sources;
    }

    public static void main( String[] args ) {
        //com.agentfactory.platform.InitialisationManager.main(null);
        AFAPL2Compiler compiler = new AFAPL2Compiler();
        for ( int i = 0; i < args.length; i++ ) {
            System.err.println( "compiling: " + args[i] );
            compiler.addTarget( args[i] );
        }

        System.exit( compiler.compileFiles() ? 0 : 1 );
    }
}
