package com.agentfactory.afapl2.compiler;

import java.util.LinkedList;
import java.util.List;

import com.agentfactory.afapl2.compiler.parser.Visitor;
import com.agentfactory.afapl2.compiler.parser.node.ActionNode;
import com.agentfactory.afapl2.compiler.parser.node.ActivityNode;
import com.agentfactory.afapl2.compiler.parser.node.FunctionNode;
import com.agentfactory.afapl2.compiler.parser.node.ModuleNode;
import com.agentfactory.afapl2.compiler.parser.node.Node;
import com.agentfactory.afapl2.compiler.parser.node.OntologyNode;
import com.agentfactory.afapl2.compiler.parser.node.PerceptorNode;
import com.agentfactory.afapl2.compiler.parser.node.PlanNode;
import com.agentfactory.afapl2.compiler.parser.node.PredicateNode;
import com.agentfactory.afapl2.compiler.parser.node.ProgramNode;
import com.agentfactory.afapl2.compiler.parser.node.TypeDefNode;

/**
 * In this phase the compiler reads ontology references, checks for the
 * existence of any specified Java files (e.g. in actions, perceptors, modules)
 * and finally builds up a list of activities.
 *
 * @author Rem Collier
 */
public class Phase2Visitor extends Visitor {
    protected List<ActivityNode> activities;
    protected List<OntologyNode> ontologies;
    
    // The following 2 fields should replace the activities field that is
    // given above.
    
    protected List<PlanNode> plans;
    protected List<ActionNode> actions;
    
    public Phase2Visitor(List<AFAPL2SyntaxError> errors, List<AFAPL2SyntaxError> warnings) {
        super(errors,warnings);
        this.activities = new LinkedList<ActivityNode>();
        plans = new LinkedList<PlanNode>();
        actions = new LinkedList<ActionNode>();
            
        ontologies = new LinkedList<OntologyNode>();
    }

    @Override
	public void visit(Node node) {
        if (node instanceof ProgramNode) {
            visit((ProgramNode) node);
        } else if (node instanceof PerceptorNode) {
            visit((PerceptorNode) node);
        } else if (node instanceof ActionNode) {
            visit((ActionNode) node);
        } else if (node instanceof PlanNode) {
            visit((PlanNode) node);
        } else if (node instanceof ModuleNode) {
            visit((ModuleNode) node);
        } else if (node instanceof OntologyNode) {
            visit((OntologyNode) node);
        } else {
            throw new RuntimeException("Attempted to visit a Node instance: " + node);
        }
    }
    
    private void visit(ProgramNode node) {
        for (PerceptorNode _node: node.getPerceptors()) {
            _node.accept(this);
        }
        for (ActionNode _node: node.getActions()) {
            _node.accept(this);
        }
        for (PlanNode _node: node.getPlans()) {
            _node.accept(this);
        }
        for (ModuleNode _node:node.getModules()) {
            _node.accept(this);
        }
        for (OntologyNode _node:node.getOntologies()) {
            _node.accept(this);
        }
    }
    
    protected void visit(PerceptorNode node) {
        try {
            Class.forName(node.getPerceptor().toString());
        } catch (ClassNotFoundException cnfe) {
            addWarning(node.getSource(), node.getLine(), "PERCEPTOR declaration references unknown Perceptor class: " + node.getPerceptor().toString());
        }
    }
        
    protected void visit(ActionNode node) {
        try {
            Class.forName(node.getActionClass().toString());
            
            // Only add to activity list if actuator exists...
            if (!activities.contains(node.getIdentifier())) {
                activities.add(node.getIdentifier());
                actions.add(node);
            }
        } catch (ClassNotFoundException cnfe) {
            addWarning(node.getSource(), node.getLine(), "ACTION declaration references unknown Actuator class: " + node.getActionClass().toString());
        }
    }
    
    protected void visit(ModuleNode node) {
        try {
            Class.forName(node.getModuleClass().toString());
        } catch (ClassNotFoundException cnfe) {
            addWarning(node.getSource(), node.getLine(), "LOAD_MODULE declaration references unknown Module class: " + node.getModuleClass().toString());
        }
    }

    private void visit(PlanNode node) {
        if (!activities.contains(node.getIdentifier())) {
            activities.add(node.getIdentifier());
            plans.add(node);
        }
    }
    
    protected void visit(OntologyNode node) {
        String identifier = node.getIdentifier();
        
        for (TypeDefNode tn: node.getTypes()) {
            if (!node.isValidType(tn.getType())) {
                addError(node.getSource(), node.getLine(), "Validation Error in Ontology " + identifier, "Unknown Type: " + tn.getType());
            }
        }
        for (PredicateNode pn: node.getPredicates()) {
            for (String type: pn.getTypes()) {
                if (!node.isValidType(type)) {
                    addError(node.getSource(), node.getLine(), "Validation Error in Ontology " + identifier, "Unknown Type in predicate " + pn.getFunctor() + ": " + type);
                }
            }
        }
        for (FunctionNode fn: node.getFunctions()) {
            for (String type: fn.getTypes()) {
                if (!node.isValidType(type)) {
                    addError(node.getSource(), node.getLine(), "Validation Error in Ontology " + identifier, "Unknown Type in function " + fn.getFunctor() + ": " + type);
                }
            }
            if (fn.getImplementation() != null) {
                try {
                    Class.forName(fn.getImplementation());
                } catch (ClassNotFoundException cnfe) {
                    addError(node.getSource(), node.getLine(), "Validation Error in Ontology " + identifier, "Unknown Class defined in function " + fn.getFunctor() + ": " + fn.getImplementation());
                }
            }
        }

        ontologies.add(node);
    }
    
    public List<PlanNode> getPlans() {
        return plans;
    }
    
    public List<ActionNode> getActions() {
        return actions;
    }
    
    public List<OntologyNode> getOntologies() {
        return ontologies;
    }
}
