/*
 * ProgramNode.java
 *
 * Created on 13 June 2006, 14:54
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package com.agentfactory.afapl2.compiler.parser.node;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import com.agentfactory.afapl2.compiler.parser.LexicalAnalyser;
import com.agentfactory.afapl2.compiler.parser.LexicalAnalyserException;
import com.agentfactory.afapl2.compiler.parser.ParseTreeException;
import com.agentfactory.afapl2.compiler.parser.ParserHelper;
import com.agentfactory.afapl2.compiler.parser.token.Token;
import com.agentfactory.afapl2.compiler.typeCheck.TypeChecker;

/**
 *
 * Syntax:
 * 
 * ONTOLOGY <fos> {
 *     SOURCE <uri>;
 * 
 *     TERM <fos>;
 *     TERM <fos>;
 * }
 * 
 * @author Rem Collier
 */
public class OntologyNode extends Node {
    private FOSNode id;
    private ClassNode source;
    private List<PredicateNode> predicates;
    private List<EnumNode> enumerations;
    private List<TypeDefNode> types;
    private List<FunctionNode> functions;
    private TypeChecker typeChecker;
    private static Set<String> basicTypes;

    public static Set<String> getBasicTypes() {
        if (basicTypes == null) {
            basicTypes = new HashSet<String>();
            basicTypes.add("short");
            basicTypes.add("int");
            basicTypes.add("long");
            basicTypes.add("float");
            basicTypes.add("double");
            basicTypes.add("char");
            basicTypes.add("boolean");
            basicTypes.add("string");
            basicTypes.add("sequence");
            basicTypes.add("set");
            basicTypes.add("none");
        }
        return basicTypes;
    }
    
    /** Creates a new instance of ProgramNode */
    public OntologyNode() {
        predicates = new LinkedList<PredicateNode>();
        enumerations = new LinkedList<EnumNode>();
        types = new LinkedList<TypeDefNode>();
        functions = new LinkedList<FunctionNode>();
    }
    
    @Override
	public void construct(LexicalAnalyser analyser) throws LexicalAnalyserException, ParseTreeException {
        Token token = analyser.getNextToken();
        id = new FOSNode(token.getText());
        
        ParserHelper.punctuationCheck("{", "Syntax Error in ONTOLOGY declaration: ONTOLOGY " + id, "expected { but got", analyser);

        token = analyser.getNextToken();
        while (!token.getText().equals("}")) {
            if (token.getText().equals("SOURCE")) {
                source = ClassNode.createClassNode(analyser.getNextToken(), analyser);
                ParserHelper.punctuationCheck(";", "Syntax Error in SOURCE parameter of ONTOLOGY " + id, "expected ; but got", analyser);
            } else if (token.getText().equals("PREDICATE")) {
                token = analyser.getNextToken();
                predicates.add(PredicateNode.create(token, analyser));
                ParserHelper.punctuationCheck(";", "Syntax Error in PREDICATE parameter of ONTOLOGY " + id, "expected ; but got", analyser);
            } else if (token.getText().equals("FUNCTION")) {
                token = analyser.getNextToken();
                functions.add(FunctionNode.create(token, analyser));
                ParserHelper.punctuationCheck(";", "Syntax Error in FUNCTION parameter of ONTOLOGY " + id, "expected ; but got", analyser);
            } else if (token.getText().equals("ENUM")) {
                token = analyser.getNextToken();
                enumerations.add(EnumNode.create(token, analyser));
                ParserHelper.punctuationCheck(";", "Syntax Error in ENUM parameter of ONTOLOGY " + id, "expected ; but got", analyser);
            } else if (token.getText().equals("TYPEDEF")) {
                token = analyser.getNextToken();
                types.add(TypeDefNode.create(token, analyser));
                ParserHelper.punctuationCheck(";", "Syntax Error in TYPEDEF parameter of ONTOLOGY " + id, "expected ; but got", analyser);
            } else {
                throw new ParseTreeException("Syntax Error in ONTOLOGY " + id, "expected SOURCE or TERM but got", token);
            }
            token = analyser.getNextToken();
        }
        
        token = analyser.getNextToken();
        if (token != null) {
            analyser.back(token);
        }
        
        if ((token == null) || !token.getText().equals(";")) {
            Token temp = Token.createToken(";", -1, null);
            analyser.back(temp);
        }
        
        typeChecker = new TypeChecker(this);
    }
    
    
    @Override
	public String toString() {
        return "ONTOLOGY " + source;
    }
    
    public String getResource() {
        if (source == null) return null;
        return source.toString();
    }
    
    public List<PredicateNode> getPredicates() {
        return predicates;
    }
    
    public List<EnumNode> getEnumerations() {
        return enumerations;
    }
    
    public List<TypeDefNode> getTypes() {
        return types;
    }
    
    public List<FunctionNode> getFunctions() {
        return functions;
    }
    
    public String getIdentifier() {
        return id.toString();
    }
    
    public boolean isValidType(String type) {
        if (type.endsWith(">")) {
            // we have a sequence or a set
            String[] check = type.substring(0, type.length()-1).split("<");
            if (check[0].equals("sequence") || check[0].equals("set")) {
                return isValidType(check[1]);
            }
        }
            
        if (getBasicTypes().contains(type)) return true;
        
        for (EnumNode node: enumerations) {
            if (node.getIdentifier().equals(type)) return true;
        }
        
        for (TypeDefNode node: types) {
            if (node.getIdentifier().equals(type)) return true;
        }
        
        for (FunctionNode node: functions) {
            if (node.getFunctor().equals(type)) return true;
        }
        
        return false;
    }
    
    public boolean validateArgument(String type, FOSNode arg) {
        return typeChecker.validateTerm(type, arg);
    }
}
