package com.agentfactory.afapl2.interpreter;

import java.io.IOException;
import java.util.List;
import java.util.Observable;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.agentfactory.afapl2.interpreter.decoder.ActionDecoder;
import com.agentfactory.afapl2.interpreter.decoder.AlwaysDecoder;
import com.agentfactory.afapl2.interpreter.decoder.BeliefDecoder;
import com.agentfactory.afapl2.interpreter.decoder.BeliefRuleDecoder;
import com.agentfactory.afapl2.interpreter.decoder.CommitmentDecoder;
import com.agentfactory.afapl2.interpreter.decoder.CommitmentRuleDecoder;
import com.agentfactory.afapl2.interpreter.decoder.FOSDecoder;
import com.agentfactory.afapl2.interpreter.decoder.ModuleDecoder;
import com.agentfactory.afapl2.interpreter.decoder.OntologyDecoder;
import com.agentfactory.afapl2.interpreter.decoder.PerceptorDecoder;
import com.agentfactory.afapl2.interpreter.decoder.PlanDecoder;
import com.agentfactory.afapl2.interpreter.decoder.RoleDecoder;
import com.agentfactory.afapl2.interpreter.mentalState.Action;
import com.agentfactory.logic.agent.Module;
import com.agentfactory.logic.agent.beliefManagement.IBeliefManagerRuleSupport;
import com.agentfactory.logic.lang.FOS;
import com.agentfactory.platform.core.IAgent;

/**
 * Implements an agent that behaves according to an AFAPL2 script, as those
 * produced by {@link com.agentfactory.afapl2.AFAPL2Compiler}.
 *
 * This class populates the data-structures of an {@code AbstractAFAPL2Agent}
 * from an AFAPL2 script.
 *
 * @author <a href="mailto:rem.collier@ucd.ie">Rem Collier</a>
 */
public class ScriptBasedAFAPL2Agent extends AbstractAFAPL2Agent {
    public ScriptBasedAFAPL2Agent(String name) {
        super(name);
    }

    /**
     * Constructs the agent by parsing an AFAPL2 script.
     *
     * @param scanner a tokenizer for AFAPL2 scripts
     */
    public void construct(Scanner scanner) {
        String keyword = null;
        List<String> statement = null;
        while (scanner.hasNext()) {
            try {
                statement = scanner.nextStatement();
                if (statement.size() > 0) {
                    try {
                        keyword = statement.get(0);
                        if (keyword.equals("ACT")) {
                            ActionDecoder decoder = new ActionDecoder(statement);
                            decoder.decode();
                            Action action = decoder.getAction(this);
                            actions.put(InterpreterHelper.createSignature(action.getIdentifier().toFOS()), action);
                        } else if (keyword.equals("PER")) {
                            PerceptorDecoder decoder = new PerceptorDecoder(statement);
                            decoder.decode();
                            perceptors.add(decoder.getPerceptor(this));
                        } else if (keyword.equals("PL")) {
                            PlanDecoder decoder = new PlanDecoder(statement);
                            decoder.decode();
                            planLibrary.addPlan(decoder.getPlan());
                        } else if (keyword.equals("COM")) {
                            CommitmentDecoder decoder = new CommitmentDecoder(statement, 1);
                            decoder.decode();
                            commitmentManager.addInitialCommitment(decoder.getCommitment());
                        } else if (keyword.equals("B")) {
                            BeliefDecoder decoder = new BeliefDecoder(statement);
                            decoder.decode();
                            beliefManager.adopt(new FOS(decoder.getBelief()));
                        } else if (keyword.equals("ONT")) {
                            OntologyDecoder decoder = new OntologyDecoder(statement);
                            decoder.decode();
                            ontologyManager.addOntology(decoder.getOntology());
                        } else if (keyword.equals("AL")) {
                            AlwaysDecoder decoder = new AlwaysDecoder(statement);
                            decoder.decode();
                            beliefManager.adopt(new FOS(decoder.getBelief()));
                        } else if (keyword.equals("G")) {
                            FOSDecoder decoder = new FOSDecoder(statement, 1);
                            decoder.decode();
                            goalManager.adoptGoal("GOAL(" + decoder.getFOS() + ")");
                        } else if (keyword.equals("R")) {
                            RoleDecoder decoder = new RoleDecoder(statement);
                            decoder.decode();
                            roleLibrary.addRoleTemplate(decoder.getRoleTemplate());
                        } else if (keyword.equals("MOD")) {
                            ModuleDecoder decoder = new ModuleDecoder(statement);
                            decoder.decode();
                            Module module = decoder.getModule(this);
                            modules.put(module.getName(), module);
                        } else if(keyword.equals("CR")) {
                            CommitmentRuleDecoder decoder = new CommitmentRuleDecoder(statement);
                            decoder.decode();
                            commitmentManager.addCommitmentRule(decoder.getRule());
                        } else if(keyword.equals("BR")) {
                            BeliefRuleDecoder decoder = new BeliefRuleDecoder(statement);
                            decoder.decode();
                            ((IBeliefManagerRuleSupport) beliefManager).addRule(decoder.getRule());
                        } else {
                            Logger.getLogger(name).log(Level.WARNING,"Unknown Keyword: " + keyword);
                        }
                    } catch (Throwable th) {
                        Logger.getLogger(name).log(Level.SEVERE, "Failed to parse input: " + statement, th);
                    }
                }
            } catch (IOException ioe) {
                Logger.getLogger(name).log(Level.SEVERE, "Problem parsing agent design for: " + name, ioe);
            }
        }
        
        state = IAgent.INITIATED;
    }

    public void update(Observable o, Object arg) {
//        throw new UnsupportedOperationException("Not supported yet.");
    }
}
