package com.agentfactory.afapl2.interpreter.control;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import com.agentfactory.afapl2.interpreter.AbstractAFAPL2Agent;
import com.agentfactory.afapl2.interpreter.IGoalManager;
import com.agentfactory.afapl2.interpreter.mentalState.Action;
import com.agentfactory.afapl2.interpreter.mentalState.Commitment;
import com.agentfactory.afapl2.interpreter.mentalState.Goal;
import com.agentfactory.afapl2.interpreter.mentalState.Plan;
import com.agentfactory.logic.lang.FOS;
import com.agentfactory.logic.lang.TerAND;
import com.agentfactory.logic.reasoner.SubstitutionSet;
import com.agentfactory.logic.reasoner.Unifier;

/**
 * Implements a goal based reasoner for AFAPL2
 * 
 * @author Rem Collier
 */
public class GoalManager implements IGoalManager {
    private AbstractAFAPL2Agent agent;
    private Set<Goal> goals;

    /**
     * Creates a new instance of GoalManager
     * 
     * @param _agent
     *            the associated agent
     */
    public GoalManager( AbstractAFAPL2Agent _agent ) {
        agent = _agent;
        goals = new HashSet<Goal>();
    }

    /**
     * Returns a list of goals
     * 
     * @return the list of goals
     */
    public Set<Goal> getGoals() {
        return goals;
    }

    /**
     * Add a new goal. The goal will only be added in the event that it is not
     * already a goal.
     * 
     * @param _goal
     *            the goal to be added
     * 
     * @return true if the goal is added, false otherwise
     */
    public Goal adoptGoal( String _goal ) {
        return adoptGoal( new Goal( _goal ) );
    }

    /**
     * Add a new goal. The goal will only be added in the event that it is not
     * already a goal.
     * 
     * @param _goal
     *            the goal to be added
     * 
     * @return true if the goal is added, false otherwise
     */
    public Goal mantainGoal( String _goal ) {
        Goal goal = new Goal( _goal );
        goal.setIsToBeMantained( true );
        return adoptGoal( goal );
    }

    public Goal adoptGoal( Goal goal ) {
        if ( goals.contains( goal ) ) {
            return null;
        }

        goals.add( goal );
        return goal;
    }

    /**
     * Drop a goal.
     * 
     * @param _goal
     *            the goal to be dropped, in String form
     * 
     * @return true if the goal is dropped, false otherwise (if the goal was not
     *         found)
     */
    public boolean dropGoal( String _goal ) {
        Goal goal = new Goal( _goal );
        Goal toRemove = null;

        // need to iterate so as to find the specific instance of this goal
        // that's attached to the GoalManager that contains any associated
        // commitment, hence not just using goals.remove( goal );
        for ( Goal g : this.goals ) {
            if ( g.equals( goal ) ) {
                toRemove = g;
                break;
            }
        }

        if ( toRemove == null )
            return false;

        // Finally, check if there is a corresponding commitment to achieve
        // the goal, and if there is, drop it by making it redundant.
        if ( goal.getCommitment() != null ) {
            goal.getCommitment().update( Commitment.EVENT_REDUNDANT );
        }

        // now we can remove the goal
        goals.remove( toRemove );

        return true;
    }

    /**
     * Required by the Queryable interface and is used to check what kind of
     * information the query source handles.
     * 
     * @param type
     *            the type of information (e.g. BELIEF, GOAL, COMMIT)
     * @return true if the type is a GOAL and false otherwise
     */
    public boolean hasType( String type ) {
        return type.equals( "GOAL" );
    }

    /**
     * Required by the Queryable interface and used to return a list of FOS
     * objects that match the given query.
     * 
     * @param fos
     *            the fos query
     * @return a list of matching fos objects
     */
    public List<Object> query( FOS fos ) {
        List<Object> list = new ArrayList<Object>();

        for ( Goal goal : goals ) {
            if ( goal.matches( fos ) ) {
                list.add( goal );
            }
        }

        return list;
    }

    private SubstitutionSet getBindingForPostCondition( TerAND post, FOS goalAsBelief ) {
        for ( FOS belief : post ) {
            if ( goalAsBelief.matches( belief ) ) {
                return Unifier.unifyFormulae( belief, goalAsBelief );
            }
        }
        return null;
    }

    public void manage() {
        for ( Iterator<Goal> i = goals.iterator(); i.hasNext(); ) {
            Goal goal = i.next();

            // Check if the goal has been realised
            if ( agent.query( goal.getCorrespondingBelief() ) ) {
                goal.setState( Goal.STATE_SUCCEEDED );
            }

            switch ( goal.getState() ) {
                case Goal.STATE_NEW:
                    FOS goalAsBelief = goal.getCorrespondingBelief();

                    // Add actions as options
                    for ( Action action : agent.getActions().values() ) {
                        SubstitutionSet set = getBindingForPostCondition( action.getPostCondition(), goalAsBelief );

                        if ( set != null ) {
                            List<SubstitutionSet> preSet = agent.queryAll( action.getPreCondition().apply( set ) );
                            if ( preSet != null ) {
                                if ( preSet.isEmpty() ) {
                                    goal.addOption( action.getIdentifier().toFOS().apply( set ) );
                                }
                                else {
                                    goal.addOption( action.getIdentifier().toFOS().apply( set ).apply( preSet.get( 0 ) ) );
                                }
                            }
                        }
                    }

                    // Add Plans as options
                    for ( Plan plan : agent.getPlanLibrary().getPlans() ) {
                        SubstitutionSet set = getBindingForPostCondition( plan.getPostCondition(), goalAsBelief );

                        if ( set != null ) {
                            List<SubstitutionSet> preSet = agent.queryAll( plan.getPreCondition().apply( set ) );
                            if ( preSet != null ) {
                                if ( preSet.isEmpty() ) {
                                    goal.addOption( plan.getIdentifier().toFOS().apply( set ) );
                                }
                                else {
                                    goal.addOption( plan.getIdentifier().toFOS().apply( set ).apply( preSet.get( 0 ) ) );
                                }
                            }
                        }
                    }

                    goal.setState( Goal.STATE_READY );
                    break;
                case Goal.STATE_READY:
                    FOS activity = goal.getNextOption();
                    if ( activity == null ) {
                        goal.setState( Goal.STATE_FAILED );
                    }
                    else {
                        TerAND condition = new TerAND();
                        condition.add( goal );
                        Commitment commitment = new Commitment( new FOS( agent.getName() ), new FOS( "?now" ), condition, activity, null );
                        agent.getCommitmentManager().addInitialCommitment( commitment );
                        goal.setState( Goal.STATE_ACTIVE );
                        goal.setCommitment( commitment );
                    }
                    break;
                case Goal.STATE_ACTIVE:
                    if ( !agent.getCommitmentManager().hasCommitment( goal.getCommitment() ) )
                        goal.setState( Goal.STATE_READY );
                    break;
                case Goal.STATE_FAILED:
                    agent.adoptBelief( new FOS( "BELIEF(unachieveableGoal(" + goal.argAt( 0 ) + "))" ) );
                case Goal.STATE_SUCCEEDED:
                    if ( goal.isToBeMantained() ) {
                        goal.reset();
                        goal.setState( Goal.STATE_NEW );
                    }
                    else {
                        i.remove();
                    }
            }
        }
    }
}
