package com.agentfactory.afapl2.interpreter.control;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.agentfactory.afapl2.interpreter.mentalState.Plan;
import com.agentfactory.afapl2.interpreter.mentalState.PlanInstance;
import com.agentfactory.logic.lang.FOS;
import com.agentfactory.logic.reasoner.SubstitutionSet;
import com.agentfactory.logic.reasoner.Unifier;

/**
 * Implement a container for all plan libraries available to the agent on the
 * current platform. Plans are firstly created and then added to an instance of
 * the plan library. For the time being plan libraries are considered to be
 * specific to the platform rather than to the agent, but this could change in
 * future.
 * 
 * Note : Polymorphic plans are not allowed. A plan is uniquely identified by
 * the functor of the plan identifier i.e. neither the number of arguements, nor
 * the type of arguements are used to identify the plan, although arguements
 * must of course be provided to initialise an instance of the plan. This is
 * analagous to a C function call.
 * 
 * @author <a href="mailto:robertr@tzi.de">Robert Ross</a>
 * 
 */
public class PlanLibrary {
    /**
     * Set the logging level of this class
     */
    private static final Logger logger = Logger.getLogger( PlanLibrary.class.getName() );

    static {
        logger.setLevel( Level.SEVERE );
    }

    /**
     * Container for different plans
     */
    private Map<String, Plan> library = null;
    private List<Plan> plans;

    /**
     * Construct an empty plan library. Plans can later be added one by one to
     * the plan library, or may be added in batch through the use of the
     * loadPlansFromConfig(String) and loadDefaultPlansFromConfig() methods.
     * Internally plans are stored in a hashtable of plans and their keys. Plan
     * keys correspond to the plan identifiers detailed in the actual plan
     * files.
     */
    public PlanLibrary() {
        this.library = new HashMap<String, Plan>();
        this.plans = new LinkedList<Plan>();
    }

    /**
     * Add a plan to the plan library. The plan's identifier will be used as the
     * key in the hashtable. If a plan by that name already exists in the
     * container, an exception will be thrown.
     * 
     * @parameter inPlan - A Plan object. A plan which is to be added to the
     *            plan library. The planIdentifier will be used as a key to
     *            store this plan.
     * @throws AgentConfigurationException
     *             - thrown if a plan with a key mathing the incomming plan is
     *             already in the library.
     */
    public void addPlan( Plan inPlan ) {
        if ( library.get( inPlan.getIdentifier().getPredicate() ) == null ) {
            this.library.put( inPlan.getIdentifier().getPredicate(), inPlan );
            plans.add( inPlan );
        }
        else {
            // else a plan with that identifier was already in there.
            // we throw an exception to indicate this.
            throw new RuntimeException( "Duplicate Plan Exists in Plan Library: " + inPlan.getIdentifier() );
        }
    }

    /**
     * Method used to retrieve and initialise plan from the plan library . In
     * practice the plan library only stores original template plans, and these
     * are initialised to specific parameters as supplied to this method.
     * 
     * @parameter inPlanIdentifier - A FOS object which identifies the plan
     *            which is to be retrieved. It is this FOS which is used to
     *            initialise the plan.
     * @returns Plan - The cloned and initialised plan instance which matches
     *          the plan call requested.
     */
    public PlanInstance getPlan( FOS pid ) {
        Plan plan = library.get( pid.getFunctor() );
        if ( plan != null ) {
            if ( pid.numArguments() != plan.getIdentifier().numArguments() ) {
                throw new RuntimeException( "initialising plan : number of arguments do not match - call was : " + pid );
            }
            SubstitutionSet subs = Unifier.unifyFormulae( pid, plan.getIdentifier().toFOS() );
            logger.log( Level.INFO, "Selecting Plan: " + pid + "\nMatching Plan: " + plan + "\nBinding: " + subs + "\nAfter Binding: " + plan.getBody().apply( subs ) );
            return new PlanInstance( plan, subs );
        }

        return null;
    }

    /**
     * Get a copy of the plan library as a hashtable. This method is put in
     * place to allow other classes to iterate thorugh the hashtable for
     * searching. Possibly not a good idea to make this available to all, but it
     * will work for the time being.
     */
    public Map<String, Plan> getLibrary() {
        return this.library;
    }

    public List<Plan> getPlans() {
        return plans;
    }
}