package com.agentfactory.astr.interpreter;

import java.io.Serializable;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.agentfactory.clf.interpreter.Action;
import com.agentfactory.clf.interpreter.CoreUtilities;
import com.agentfactory.clf.interpreter.Module;
import com.agentfactory.clf.interpreter.Sensor;
import com.agentfactory.clf.lang.IMPLIES;
import com.agentfactory.clf.lang.Predicate;
import com.agentfactory.clf.plans.AssignmentStep;
import com.agentfactory.clf.plans.BlockStep;
import com.agentfactory.clf.plans.DurativeStep;
import com.agentfactory.clf.plans.ForEachStep;
import com.agentfactory.clf.plans.IPlanStep;
import com.agentfactory.clf.plans.IfStep;
import com.agentfactory.clf.plans.TryRecoverStep;
import com.agentfactory.clf.plans.WaitStep;
import com.agentfactory.clf.plans.WhileStep;
import com.agentfactory.service.migration.IMigrationSnapshot;

public class ASTRMigrationSnapshot implements IMigrationSnapshot,Serializable {
	private static final long serialVersionUID = -4736795445676840945L;
	
	public class Rule implements Serializable {
		private static final long serialVersionUID = 1L;
		
		String label;
		String event;
		String context;
		List<String> plan;
		
		public Rule(PlanRule rule) {
			label = Utilities.presenter.toString(rule.getLabel());
			event = rule.getTriggeringEvent().toString();
			context = Utilities.presenter.toString(rule.getContext());
			plan = new LinkedList<String>();
			for (IPlanStep step : rule.getPlan()) {
				plan.add(createPlanStep(step));
			}
		}
		
		private String createPlanStep(IPlanStep step) {
			if (ASTRBasicStep.class.isInstance(step)) {
				return "BS " + convertSpaces(Utilities.presenter.toString(((ASTRBasicStep)step).formula()));
			} else if (AssignmentStep.class.isInstance(step)) {
				AssignmentStep _step = (AssignmentStep) step;
				return "AS " + Utilities.presenter.toString(_step.variable()) + " " + convertSpaces(Utilities.presenter.toString(_step.expression()));
			} else if (WhileStep.class.isInstance(step)) {
				WhileStep _step = (WhileStep) step;
				return "WH " + convertSpaces(Utilities.presenter.toString(_step.guard())) + " " + createPlanStep(_step.plan());  
			} else if (ForEachStep.class.isInstance(step)) {
				ForEachStep _step = (ForEachStep) step;
				return "FE " + convertSpaces(Utilities.presenter.toString(_step.guard())) + " " + createPlanStep(_step.body());  
			} else if (DurativeStep.class.isInstance(step)) {
				DurativeStep _step = (DurativeStep) step;
				return "DU " + convertSpaces(Utilities.presenter.toString(_step.formula()));  
			} else if (IfStep.class.isInstance(step)) {
				IfStep _step = (IfStep) step;
				return "IF " + convertSpaces(Utilities.presenter.toString(_step.guard()))+ " " + createPlanStep(_step.ifStep()) + " " + createPlanStep(_step.elseStep());
			} else if (TryRecoverStep.class.isInstance(step)) {
				TryRecoverStep _step = (TryRecoverStep) step;
				return "TR " + createPlanStep(_step.plan()) + " " + createPlanStep(_step.recoverPlan());
			} else if (WaitStep.class.isInstance(step)) {
				WaitStep _step = (WaitStep) step;
				return "WT " + convertSpaces(Utilities.presenter.toString(_step.guard()));
			} else if (BlockStep.class.isInstance(step)) {
				BlockStep _step = (BlockStep) step;
				StringBuffer buf = new StringBuffer();
				buf.append("SB");
				for (IPlanStep stp: _step.plan()) {
					buf.append(" " + createPlanStep(stp));
				}
				return buf.toString() + " EB";
			}
			
			System.out.println("Unknown PlanStep: " + step.getClass().getName());
			return null;
		}
		
		private String convertSpaces(String str) {
			return str.replace(" ", "%20");
		}
	}
	
	String name;
	List<String> beliefs;
	List<String> events;
	Map<String,String> modules;
	Map<String,String> actions;
	Map<String,String> sensors;
	List<Rule> rules;
	List<String> inferences;
	
	public ASTRMigrationSnapshot(AbstractASTRAgent agent) {
		name = agent.getName();
		
		beliefs = new LinkedList<String>();
		for (Predicate predicate: agent.getBeliefs()) {
			beliefs.add(Utilities.presenter.toString(predicate));
		}
		inferences = new LinkedList<String>();
		for (IMPLIES inference:agent.getInferences()) {
			inferences.add(CoreUtilities.presenter.toString(inference));
		}
		
		events = new LinkedList<String>();
		for (ASEvent event: agent.getEventQueue()) {
			events.add(event.toString());
		}
		rules = new LinkedList<Rule>();
		for (PlanRule rule : agent.getRules()) {
			rules.add(new Rule(rule));
		}
		modules = new HashMap<String, String>();
		for (Module module: agent.getModuleMap().values()) {
			modules.put(Utilities.presenter.toString(module.getId()), module.getClass().getName());
		}
		actions = new HashMap<String, String>();
		for (Action act: agent.getActionMap().values()) {
			actions.put(Utilities.presenter.toString(act.getId()), act.getClass().getName());
		}
		sensors = new HashMap<String, String>();
		for (Sensor sensor: agent.getSensorMap().values()) {
			actions.put(Utilities.presenter.toString(sensor.getId()), sensor.getClass().getName());
		}
	}
	
	@Override
	public String getName() {
		return name;
	}

	@Override
	public String getType() {
		return "AgentSpeak";
	}
}
