package com.agentfactory.clf.interpreter;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.agentfactory.clf.lang.Function;
import com.agentfactory.clf.lang.IFormula;
import com.agentfactory.clf.lang.ITerm;
import com.agentfactory.clf.lang.ListTerm;
import com.agentfactory.clf.lang.Predicate;
import com.agentfactory.clf.reasoner.Bindings;
import com.agentfactory.clf.reasoner.QueryResult;
import com.agentfactory.platform.core.AgentID;
import com.agentfactory.platform.core.IMessage;
import com.agentfactory.platform.core.IPlatformService;
import com.agentfactory.platform.impl.AbstractAgent;
import com.agentfactory.platform.impl.StringMessage;
import com.agentfactory.platform.util.FIPAHelper;

public abstract class AbstractLanguageAgent extends AbstractAgent {
	protected class MigrationState {
		public static final int NONE						= -1;
		public static final int IMMEDIATE 					= 0;
		public static final int GRACEFULL					= 1;
		
		public boolean migrate = false;
		public int type = NONE;
		public String host;
		public int port;
	}
	
	protected Map<String, Action> actionMap;
	protected Map<String, Sensor> sensorMap;
	protected Map<String, Module> moduleMap;
	protected List<String> actionLog;
	protected MigrationState migrationState;
	
	public AbstractLanguageAgent(String name) {
		super(name);
		
		actionLog = new LinkedList<String>();
		setActionMap(new HashMap<String, Action>());
		setSensorMap(new HashMap<String, Sensor>());
		moduleMap = new HashMap<String, Module>();
		migrationState = new MigrationState();
	}
	
	/**
	 * Default Sensing Algorithm:
	 * 
	 * 1. agent level sensing
	 * 2. module level sensing
	 * 
	 */
	protected void senseEnvironment() { 
		// Part A: Sense the environment...
		for (Sensor sensor: getSensorMap().values()) {
			sensor.perceive();
		}
        for (Module module: moduleMap.values()) {
    		for (Sensor sensor: module.getSensors()) {
    			sensor.perceive();
    		}
        }
	}

	public void noSuchAction(Predicate activity) {
	}
	
	/**
	 * Invoke the specified action:
	 * 
	 * 1. attempt an agent level action
	 * 2. attempt a module level action
	 * 3. if no action, then call the customisable noSuchAction() method
	 * 
	 * @param activity
	 * @return
	 */
	public boolean performActivity(Predicate activity) {
		Action action = getActionMap().get(CoreUtilities.createSignature(activity));
		if (action == null) {
			action = getModuleLevelAction(activity);
			if (action == null) {
				noSuchAction(activity);
				return false;
			}
		}
		
		return action.execute(activity);
	}

	/**
	 * Retrieves a module level action or returns null if none
	 * exists.
	 * 
	 * @param activity
	 * @return
	 */
	public Action getModuleLevelAction(Predicate activity) {
		String sig = CoreUtilities.createSignature(activity);
		int index = sig.indexOf(".");
		if (index == -1) return null;
		
		String id = sig.substring(0, index) +":0";
		Module module = moduleMap.get(id);
		if (module == null) {
			return null;
		}
		return module.getAction(sig.substring(index+1));
	}

	public void addModule(String id, Module module) {
		addModule((Predicate) CoreUtilities.factory.createFormula(id), module);
	}
	
	public void addModule(Predicate id, Module module) {
		module.setId(id);
		module.setAgent(this);
		moduleMap.put(CoreUtilities.createSignature(id), module);
	}
	
	protected void addAction(String name, Action action) {
		action.setId((Predicate) CoreUtilities.factory.createFormula(name));
		action.setAgent(this);
		getActionMap().put(CoreUtilities.createSignature(action.getId()), action);
	}
	
	protected void addSensor(String name, Sensor sensor) {
		sensor.setId((Predicate) CoreUtilities.factory.createFormula(name));
		sensor.setAgent(this);
		getSensorMap().put(CoreUtilities.createSignature(sensor.getId()), sensor);
	}
	
	public Module getModule(Predicate module) {
		return moduleMap.get(CoreUtilities.createSignature(module));
	}
	
	public abstract void addBelief(Predicate belief); 
	public abstract void removeBelief(Predicate belief);

	protected void installBasicAPI() {
		addAction(".abolish(?belief)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				removeBelief(termAsString(activity, 0));
				return true;
			}
		});
		addAction(".abolishAll(?belief)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				Predicate p = CoreUtilities.factory.convertToPredicate(activity.termAt(0));
				QueryResult result = agent.queryAll(p);
				if (result.found) {
//					System.out.println("match");
					for (Bindings b: result.bindings) {
//						System.out.println("removing: " + CoreUtilities.presenter.toString(p.apply(b)));
						agent.removeBelief((Predicate) p.apply(b));
					}
				}
				return true;
			}
		});
		addAction(".send(?perf, ?rec, ?content)", new Action() {
			/**
			 * Implementation of basic send functionality: supports
			 * sending to a single agent or a list of agents.
			 */
			@Override
			public boolean execute(Predicate activity) {
				String performative = termAsString(activity, 0);
				String content = termAsString(activity, 2);
	
		        StringMessage message = StringMessage.newInstance();
		        message.setPerformative(performative.toUpperCase());
		        message.setSender(agent.getAgentID());
				if (ListTerm.class.isInstance(activity.termAt(1))) {
					for (ITerm term: activity.termAt(1).terms()) {
						message.getReceivers().add(toAgentID(term));
					}
				} else {
					message.getReceivers().add(toAgentID(activity.termAt(1)));
				}
		        message.setContent(content);
		        message.setLanguage("AFAPL");
		        message.setOntology("none");
		        message.setProtocol("none");
		        
		        return agent.send(message);
			}
			
			private AgentID toAgentID(ITerm term) {
		    	if (!Function.class.isInstance(term)) return null;
		    	
		    	Function fos = (Function) term;
		    	String id = termAsString(fos, 0);
		    	List<String> addresses = new LinkedList<String>();
		    	Function funct = (Function) fos.termAt(1);
		    	for (int i = 0; i < funct.terms().size(); i++) {
		    		addresses.add(termAsString(funct, i));
		    	}
		    	
		    	return new AgentID(id, addresses, new LinkedList<String>());
			}
		});
		addAction(".print(?content)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				System.out.print(termAsString(activity, 0));
				return true;
			}
		});
		addAction(".println(?content)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				System.out.println(termAsString(activity, 0));
				return true;
			}
		});
		addAction(".fail", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				return false;
			}
		});
		addAction(".nil", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				return true;
			}
		});
		addAction(".sleep(?time)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				long time = Long.parseLong(termAsString(activity, 0));
				try {
					Thread.sleep(time);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				return true;
			}
		});
		addAction(".bind(?id)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				String id = termAsString(activity, 0);
				if (agent.bindToPlatformService(id)) {
					this.addBelief("boundToService(" + id + ")");
					return true;
				}
				addBelief("bindToServiceFailed(" + id + ")");
				return false;
			}
		});
		addAction(".createModule(?id, ?design)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				Predicate id = CoreUtilities.factory.convertToPredicate(activity.termAt(0));
				String design = termAsString(activity, 1);
				try {
					Class<?> cl = Class.forName(design);
					Module module = (Module) cl.newInstance();
					addModule(id, module);
				} catch (InstantiationException e) {
					return false;
				} catch (IllegalAccessException e) {
					return false;
				} catch (ClassNotFoundException e) {
					return false;
				}
				return true;
			}
		});
		addAction(".migrate(?host, ?port)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				migrationState.migrate = true;
				migrationState.type=MigrationState.GRACEFULL;
				migrationState.host=termAsString(activity, 0);
				migrationState.port=termAsInt(activity, 1);
				return true;
			}
		});
		addAction(".migrateNow(?host, ?port)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				migrationState.migrate = true;
				migrationState.type=MigrationState.IMMEDIATE;
				migrationState.host=termAsString(activity, 0);
				migrationState.port=termAsInt(activity, 1);
				return true;
			}
		});
		addSensor(".services", new Sensor() {
			@Override
			public void perceive() {
				for(IPlatformService p :agent.getBoundPlatformServices()){
					addBelief("service("+p.getName()+")");
				};			
			}
		});
		addSensor(".inbox", new Sensor() {
			@Override
			public void perceive() {
		        for (IMessage message: agent.getInbox()) {
//			        System.out.println("RECEIVING: " + message);
		        	addBelief("message(" + message.getPerformative().toLowerCase() + "," +
		        			FIPAHelper.toFOSString(message.getSender()) + "," +
		        				  message.getContent() + ")");
		        }
			}
		});
		addSensor(".self", new Sensor() {
			@Override
			public void perceive() {
		        addBelief("name(" + agent.getName() + ")");
		        addBelief(FIPAHelper.toFOSString(agent.getAgentID()));
		        Calendar cal = Calendar.getInstance();
		        SimpleDateFormat sdf = new SimpleDateFormat("hh:mm:ss");
		        String[] bits = sdf.format(cal.getTime()).split(":");
		        
		        addBelief("time(" + bits[0] + "," + bits[1] + "," + bits[2] + ")");
//		        addBelief("iteration(" + this.)
			}
		});
	}
	
	public abstract QueryResult queryAll(IFormula formula);

	public void setActionMap(Map<String, Action> actionMap) {
		this.actionMap = actionMap;
	}

	public Map<String, Action> getActionMap() {
		return actionMap;
	}

	public void setSensorMap(Map<String, Sensor> sensorMap) {
		this.sensorMap = sensorMap;
	}

	public Map<String, Sensor> getSensorMap() {
		return sensorMap;
	}
	
	public Map<String, Module> getModuleMap() {
		return moduleMap;
	}
	
	public void logAction(String action) {
		actionLog.add(this.getIterations() + ": " + action);
	}
	
	public List<String> getActionLog() {
		return actionLog;
	}
	public Collection<Action> getActions() {
		List<Action> actions = new LinkedList<Action>();
		actions.addAll(actionMap.values());
		for (Module module : moduleMap.values()) {
			actions.addAll(module.getActions());
		}
		return actions;
	}

	public Collection<Sensor> getSensors() {
		List<Sensor> sensors = new LinkedList<Sensor>();
		sensors.addAll(sensorMap.values());
		for (Module module : moduleMap.values()) {
			sensors.addAll(module.getSensors());
		}
		return sensors;
	}
}
