package com.agentfactory.clf.interpreter;

import java.util.List;

import com.agentfactory.clf.lang.ITerm;
import com.agentfactory.clf.lang.ListTerm;
import com.agentfactory.clf.lang.Predicate;
import com.agentfactory.clf.lang.Variable;
import com.agentfactory.clf.reasoner.Bindings;

public class ListUtils extends Module {
	public ListUtils() {
		addAction("add(?list, ?index, ?value)", new ActiveAction() {

			/**
			 * This action takes a list, index, and value as input, and
			 * updates a single variable to be assigned to a list that
			 * is the same as the list given as the first argument, but
			 * with the value added at the specified index.
			 * 
			 * The index should be a number, but also 2 special values are
			 * supported: first = index 0; last = index [list-size]-1 
			 */
			@Override
			public boolean execute(Predicate activity,
					List<Variable> variables, Bindings bindings) {
				Variable result = variables.get(0);
				ListTerm list = (ListTerm) activity.termAt(0);
				ITerm value = activity.termAt(2);

				int index = getIndex(list, termAsString(activity, 1));
				list.terms().add(index, value);
				
				bindings.addBinding(result, list);
				return true;
			}
			
		});
		
		addAction("remove(?list, ?index)", new ActiveAction() {

			/**
			 * This action removes an item at a given list at a given index. The first
			 * variable associated with this action holds the updated list, and the second
			 * value is used to hold the value that is removed.
			 * 
			 * The index should be a number, but also 2 special values are
			 * supported: first = index 0; last = index [list-size]-1 
			 */
			@Override
			public boolean execute(Predicate activity,
					List<Variable> variables, Bindings bindings) {
				Variable variable = variables.get(0);		// variable to which the updated list is assigned
				Variable result = null;						// variable to which the removed value is assigned
				
				ListTerm list = (ListTerm) activity.termAt(0);
				int index = getIndex(list, termAsString(activity, 1));
				
				ITerm term = list.terms().remove(index);
				bindings.addBinding(variable, list);
				if (variables.size() > 1) {
					result = variables.get(1);
					bindings.addBinding(result, term);
				}
				return true;
			}
			
		});
		
		addAction("get(?list, ?index)", new ActiveAction() {

			@Override
			public boolean execute(Predicate activity,
					List<Variable> variables, Bindings bindings) {
				Variable variable = variables.get(0);		// variable to hold the value at the given index
				
				ListTerm list = (ListTerm) activity.termAt(0);
				int index = getIndex(list, termAsString(activity, 1));
				
				ITerm term = list.terms().get(index);
				bindings.addBinding(variable, term);
				return true;
			}
			
		});

		addAction("replace(?list, ?index, ?item)", new ActiveAction() {

			@Override
			public boolean execute(Predicate activity,
					List<Variable> variables, Bindings bindings) {
				Variable variable = variables.get(0);		// variable to hold the value at the given index
				
				ListTerm list = (ListTerm) activity.termAt(0);
				ITerm item = activity.termAt(2);
				
				int index = getIndex(list, termAsString(activity, 1));
				
				ITerm term = list.terms().set(index, item);
				bindings.addBinding(variable, term);
				return true;
			}
			
		});

		addAction("contains(?list, ?item)", new ActiveAction() {

			@Override
			public boolean execute(Predicate activity,
					List<Variable> variables, Bindings bindings) {
				Variable variable = variables.get(0);		// variable to hold the value at the given index
				
				ListTerm list = (ListTerm) activity.termAt(0);
				ITerm item = activity.termAt(2);
				
				int index = getIndex(list, termAsString(activity, 1));
				
				ITerm term = list.terms().set(index, item);
				bindings.addBinding(variable, term);
				return true;
			}
			
		});
	}
	
	private int getIndex(ListTerm list, String index) {
		if (index.equals("first")) {
			return 0;
		} else if (index.equals("last")) {
			return list.size() -1;
		}
		return Integer.parseInt(index);
	}
}
