package com.agentfactory.clf.lang;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import com.agentfactory.clf.reasoner.Bindings;

public class Predicate implements IFormula, Iterable<ITerm> {
	protected String functor;
	protected List<ITerm> terms;
	protected List<Predicate> labels;
	
	public Predicate(String functor) {
		this(functor, new LinkedList<ITerm>(), new LinkedList<Predicate>());
	}
	
	public Predicate(String functor, List<ITerm> terms) {
		this(functor, terms, new LinkedList<Predicate>());
	}
	
	public Predicate(String functor, List<ITerm> terms, List<Predicate> labels) {
		this.functor = functor;
		this.terms = terms;
		this.labels = labels;
	}
	
	@Override
	public IFormula apply(Bindings set) {
		
		List<ITerm> list = new LinkedList<ITerm>();
		List<Predicate> labs = new LinkedList<Predicate>();
		
		for (ITerm term: terms) {
			list.add(term.apply(set));
		}
		
		for (Predicate pred: labels) {
			labs.add((Predicate) pred.apply(set));
		}
		
		return new Predicate(functor, list, labs);
	}
	
	public ITerm termAt(int i) {
		return terms.get(i);
	}
	
	public int size() {
		return terms.size();
	}

	public String functor() {
		return functor;
	}
	
	public boolean isConstant() {
		return terms.isEmpty();
	}

	@Override
	public Iterator<ITerm> iterator() {
		return terms.iterator();
	}

	@Override
    public boolean equals(Object obj) {
    	if (Predicate.class.isInstance(obj)){
    		Predicate predicate = (Predicate) obj;
    		if (!functor.equals(predicate.functor)) return false;
    		if (terms.size() != predicate.size()) return false;
    		
    		for (int i=0; i< terms.size(); i++) {
    			if (!terms.get(i).equals(predicate.terms.get(i))) return false;
    		}
    		
    		return true;
    	}
    	return false;
    }
    
	/**
	 * Delegate to the result formula to verify type
	 */
	public String type() {
		return "raw-logic";
	}

	@Override
	public Set<Variable> getVariables() {
		Set<Variable> set = new HashSet<Variable>();
		for (ITerm term: terms) {
			set.addAll(term.getVariables());
		}
		for (Predicate label: labels) {
			set.addAll(label.getVariables());
		}
		return set;
	}
	
	@Override
	public List<IFormula> asList() {
		List<IFormula> list = new LinkedList<IFormula>();
		list.add(this);
		return list;
	}
	
	public boolean isTrue() {
		return functor.equalsIgnoreCase("true");
	}

	public boolean isFalse() {
		return functor.equalsIgnoreCase("false");
	}

	public boolean matches(IFormula formula) {
		if (Predicate.class.isInstance(formula)) {
			Predicate predicate = (Predicate) formula;
			return functor.equals(predicate.functor) &&
					(terms.size() == predicate.size());
		}
		return false;
	}

	public List<Predicate> getLabels() {
		return labels;
	}
	
	public List<ITerm> terms() {
	   return this.terms;
	}
}