package com.agentfactory.logic.reasoner;

import com.agentfactory.logic.lang.FOS;
import com.agentfactory.logic.lang.LogicFactory;
import com.agentfactory.logic.lang.TerAND;
import com.agentfactory.logic.lang.TerImplication;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This is the core component that performs resolution based reasoning over a
 * set of sources. Also, this class includes implementations for various
 * operators:
 * <ul>
 * <li>LESS_THAN
 * <li>GREATER_THAN
 * <li>EQUALS
 * <li>IN_RANGE
 * <li>CONTAINS
 * </ul>
 * 
 * @author <a href="mailto:rem.collier@ucd.ie">Rem Collier</a>
 */
public class QueryEngine {
    private static final Logger logger = Logger.getLogger( QueryEngine.class.getName() );

    static {
        logger.setLevel( Level.SEVERE );
    }

    private List<Queryable> sources;

    /** Creates a new instance of QueryEngine */
    public QueryEngine() {
        sources = new LinkedList<Queryable>();
    }

    public void addSource( Queryable source ) {
        sources.add( source );
    }

    public static final int MAX_LEVELS = 20;

    private static int counter = 0;

    // These data members are used for each query
    private List<StackEntry> completedSet;
    private List<StackEntry> possibleSolutions;
    private List<FOS> variableSet;

    private List<Object> getMatchingFormula( FOS formula ) {
        List<Object> formulae = new ArrayList<Object>();
        for ( Queryable source : sources ) {
            logger.log( Level.INFO, "Source: " + source );

            if ( source.hasType( formula.getFunctor() ) ) {
                logger.log( Level.INFO, "has type" );
                formulae.addAll( source.query( formula ) );
            }
        }
        return formulae;
    }

    /**
     * Add all the variables in a formula to this Reasoner's list of variables
     * that are to be resolved. The list of varible will look something like
     * ?var1, ?var2 .... where ?var1 etc. are FOS that start with a ? and have
     * no arguements.
     * 
     * If the query being made has no free variables, then the variable set will
     * be mpty.
     * 
     * @param inGoal
     *            the formula to be analysed. This must be a FOS or TerAND
     *            object.
     */
    private void buildVariableSetFor( Object object ) {
        if ( object instanceof TerAND ) {
            TerAND and = (TerAND) object;
            for ( FOS formula : and ) {
                buildVariableSetFor( formula );
            }
        }
        else if ( object instanceof FOS ) {
            FOS fos = (FOS) object;
            if ( fos.isVariable() ) {
                variableSet.add( fos );
            }
            else if ( fos.isComposite() ) {
                for ( FOS formula : fos ) {
                    buildVariableSetFor( formula );
                }
            }
        }
        else {
            throw new RuntimeException( "Should not reach here." );
        }
    }

    private void setup( Object object ) {
        completedSet = new ArrayList<StackEntry>();
        possibleSolutions = new ArrayList<StackEntry>();
        possibleSolutions.add( new StackEntry( object ) );
        variableSet = new ArrayList<FOS>();
        buildVariableSetFor( object );
    }

    public boolean query( Object goal ) {
        setup( goal );

        // Check the trivial "true" and "false" cases
        if ( goal instanceof FOS ) {
            FOS formula = (FOS) goal;
            if ( formula.isTrue() ) {
                return true;
            }
            else if ( formula.isFalse() ) {
                return false;
            }
        }

        // Go through all possibilities until the solution stack is empty, or
        // until a result was found.
        StackEntry solution;
        while ( completedSet.isEmpty() && ( !possibleSolutions.isEmpty() ) ) {
            // Get the next Stack Entry from the solution stack
            solution = (StackEntry) possibleSolutions.remove( 0 );

            if ( solution.goalStack.isEmpty() ) {
                // if the goal stack is empty - what we just took off the
                // solution
                // stack must be the result
                // Only one element is ever added to 'completed', but we will
                // leave the design as it is, since it is easy to compare with
                // 'resolveAll'.
                completedSet.add( solution );
            }
            else {
                // else we need to try to solve against that solution
                solve( solution );
            }
        }
        for ( int i = 0; i < completedSet.size(); i++ ) {
            ( (StackEntry) completedSet.get( i ) ).update( variableSet );
        }
        return !( completedSet.isEmpty() );
    }

    /**
     * Take a set of current beliefs and belief rules, and resolve all the
     * beliefs that are entailed.
     * 
     * The only difference between this and the above, is two thirds in.
     * 
     * @param goal
     *            - A Query Object. That which is to be checked for entailment.
     * @return boolean - true if inBeliefManager entails inQuery, false
     *         otherwise.
     */
    public boolean queryAll( Object goal ) {
        StackEntry solution;
        // Setup the reasoner -- when complete, possibleSolutions will have one
        // entry with the goal and no substitution.
        setup( goal );

        // Check the trivial "true" and "false" cases
        if ( goal instanceof FOS ) {
            FOS formula = (FOS) goal;
            if ( formula.isTrue() ) {
                return true;
            }
            else if ( formula.isFalse() ) {
                return false;
            }
        }

        // Go through possibilities until the solution stack is empty.
        while ( !possibleSolutions.isEmpty() ) {
            solution = (StackEntry) possibleSolutions.remove( 0 );
            logger.log( Level.INFO, "Checking: " + solution );

            if ( solution.goalStack.isEmpty() ) {
                logger.log( Level.INFO, "Completed" );
                // The potential solution has no goals left to be solved. This
                // potential solution must be a valid solution.
                completedSet.add( solution );
            }
            else if ( solution.level < MAX_LEVELS ) {
                // lets take this potential solution and try to work it out
                // If this is successful, a stack entry with no goals will be
                // added to the list of possible solutions. Else, if something
                // needs further investigation, a stack entry with goals as well
                // as a subsitution set will be added to the set of
                // possibilities.
                // Else, this solution was a dead end and nothing is added to
                // the set of possibilities.

                // Also, we check here whether the maximum search depth
                // (MAX_LEVELS)
                // has been reached - if so, then drop the potential solution...
                logger.log( Level.INFO, "About to call solve()" );
                solve( solution );
            }
        }

        for ( int i = 0; i < this.completedSet.size(); i++ ) {
            ( (StackEntry) completedSet.get( i ) ).update( variableSet );
        }

        return !( completedSet.isEmpty() );
    }

    /**
     * Get the set of substitutions for a formula that has been resolved.
     * 
     * This method is normally called by a Belief Manager object following the
     * successful resolving of a formula.
     * 
     * @return A set of SubstitutionSet objects that hold the valid
     *         substitutions for resoleved formula.
     */
    public List<SubstitutionSet> getSubstitutionSets() {
        SubstitutionSet set = null;
        List<SubstitutionSet> sets = new ArrayList<SubstitutionSet>();
        for ( int i = 0; i < completedSet.size(); i++ ) {
            set = ( (StackEntry) completedSet.get( i ) ).solution;
            if ( !sets.contains( set ) )
                sets.add( set );
        }
        return sets;
    }

    private boolean comparisonCheck( FOS goal ) {
        String operator = goal.getFunctor();

        if ( operator.equals( "EQUAL" ) ) {
            return goal.argAt( 0 ).equals( goal.argAt( 1 ) );
        }
        else if ( operator.equals( "LESS_THAN" ) ) {
            try {
                return Integer.parseInt( goal.argAt( 0 ).toString() ) < Integer.parseInt( goal.argAt( 1 ).toString() );
            }
            catch ( NumberFormatException nfe ) {
                return false;
            }
        }
        else if ( operator.equals( "GREATER_THAN" ) ) {
            try {
                return Integer.parseInt( goal.argAt( 0 ).toString() ) > Integer.parseInt( goal.argAt( 1 ).toString() );
            }
            catch ( NumberFormatException nfe ) {
                return false;
            }
        }
        else if ( operator.equals( "IN_RANGE" ) ) {
            try {
                int target = Integer.parseInt( goal.argAt( 0 ).toString() );
                int lower = Integer.parseInt( goal.argAt( 1 ).toString() );
                int upper = Integer.parseInt( goal.argAt( 2 ).toString() );
                return ( target >= lower ) && ( target <= upper );
            }
            catch ( NumberFormatException nfe ) {
                return false;
            }
        }
        else if ( operator.equals( "CONTAINS" ) ) {
            FOS target = goal.argAt( 0 );
            for ( FOS value : goal.argAt( 1 ).getArguments() ) {
                if ( value.matches( target ) ) {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * Apply a resolution step to the current solution
     * 
     * @param solution
     */
    private void solve( StackEntry solution ) {
        SubstitutionSet subSet = null;
        SubstitutionSet subSet2 = null;
        StackEntry stackEntry;

        // Reform a fully formed goal with applications of whatever
        // substitutions are appropriate
        FOS goal = solution.getNextGoal();
        logger.log( Level.INFO, "Resolving goal: " + goal );

        if ( goal.isNot() ) {
            if ( !( this.spawn().query( removeNot( goal ) ) ) ) {
                possibleSolutions.add( solution );
            }
        }
        else if ( goal.isTrue() ) {
            possibleSolutions.add( solution );
        }
        else if ( !goal.isFalse() ) {
            if ( comparisonCheck( goal ) ) {
                possibleSolutions.add( solution.copy( new SubstitutionSet() ) );
            }
            else {
                List<Object> matchingFormulae = this.getMatchingFormula( goal );
                logger.log( Level.INFO, "program: " + matchingFormulae );
                for ( Object nextMatch : matchingFormulae ) {
                    logger.log( Level.INFO, "Comparing with: " + nextMatch );
                    // Check what type of object this is and process
                    // accordingly.
                    if ( TerImplication.class.isInstance( nextMatch ) ) {
                        TerImplication implication = (TerImplication) nextMatch;
                        subSet = variableRenamingFor( implication );
                        logger.log( Level.INFO, "Variable Renaming: " + subSet );
                        subSet2 = Unifier.unifyFormulae( goal, implication.getRight().apply( subSet ) );
                        logger.log( Level.INFO, "Binding: " + subSet2 );

                        // if subs2 != null, we were able to unify the head of
                        // the
                        // belief rule with the current goal.
                        if ( subSet2 != null ) {
                            // make a copy of the main Stack Entry adding the
                            // tail
                            // of the belief rule -- giving us more to work out
                            stackEntry = solution.copy( subSet2 );
                            logger.log( Level.INFO, "adding goal: " + implication.getLeft().apply( subSet ).apply( subSet2 ) );
                            stackEntry.addGoal( implication.getLeft().apply( subSet ).apply( subSet2 ) );
                            possibleSolutions.add( stackEntry );
                        }
                    }
                    else if ( FOS.class.isInstance( nextMatch ) ) {
                        logger.log( Level.INFO, "Unify formulae" );
                        subSet = Unifier.unifyFormulae( goal, (FOS) nextMatch );
                        logger.log( Level.INFO, "Binding: " + subSet );
                        if ( subSet != null ) {
                            StackEntry se = solution.copy( subSet );
                            possibleSolutions.add( se );
                        }
                    }
                    else {
                        throw new RuntimeException( "Fatal Error in Query Engine: attempt to match: " + nextMatch );
                    }
                }
            }
        }
    }

    private QueryEngine spawn() {
        QueryEngine engine = new QueryEngine();
        for ( int i = 0; i < sources.size(); i++ ) {
            engine.addSource( (Queryable) sources.get( i ) );
        }
        return engine;
    }

    private FOS removeNot( FOS formula ) {
        FOS fos = null;
        try {
            fos = (FOS) LogicFactory.create( formula.toString().substring( 1 ) );
        }
        catch ( Exception e ) {
            System.out.println( "[QueryEngine.removeNot] Error removing '!' from: " + formula.toString() );
        }
        return fos;
    }

    /***************************************************************************
     * STATIC VARIABLE RENAMING METHODS
     ***************************************************************************/
    /**
     * Renames a given variable by appending the current value of the counter to
     * the variable.
     * 
     * @param formula
     *            the variable to be renamed
     * @return a new FOS object containing the renamed variable.
     */
    private static FOS renameFormula( FOS formula ) {
        return new FOS( formula.getFunctor() + counter, new ArrayList<FOS>() );
    }

    /**
     * Generates a SubstitutionSet that contains a set of bindings for variable
     * renaming.
     * 
     * @param object
     *            the formula to be renamed.
     * @return a substitution set containing the renamings.
     */
    private static SubstitutionSet variableRenamingFor( Object object ) {
        SubstitutionSet subs = new SubstitutionSet();
        privateVariableRenamingFor( object, subs );
        counter++;
        return subs;
    }

    /**
     * Recursively called method that generates the Substitution Set.
     * 
     * @param object
     *            the formula to be renamed.
     * @param subSet
     *            the substitution set to be augmented
     */
    private static void privateVariableRenamingFor( Object object, SubstitutionSet subSet ) {
        // First deal with FOS objects
        if ( object instanceof FOS ) {
            FOS formula = (FOS) object;
            if ( formula.isVariable() ) {
                if ( !subSet.hasVariable( formula ) ) {
                    subSet.addBinding( new Substitution( formula, renameFormula( formula ) ) );
                }
            }
            else if ( formula.isComposite() ) {
                for ( int i = 0; i < ( formula.getArguments() ).size(); i++ ) {
                    privateVariableRenamingFor( formula.argAt( i ), subSet );
                }
            }
        }

        // Now deal with TerAND objects
        if ( object instanceof TerAND ) {
            TerAND and = (TerAND) object;
            for ( FOS formula : and ) {
                privateVariableRenamingFor( formula, subSet );
            }
        }

        // Next its the turn of TerImplication objects
        if ( object instanceof TerImplication ) {
            TerImplication formula = (TerImplication) object;
            privateVariableRenamingFor( formula.getLeft(), subSet );
            privateVariableRenamingFor( formula.getRight(), subSet );
        }
    }

    class StackEntry {
        ArrayList<FOS> goalStack;
        SubstitutionSet solution;
        int level;

        private StackEntry() {
            goalStack = new ArrayList<FOS>();
            solution = new SubstitutionSet();
            level = 0;
        }

        public StackEntry( Object goal ) {
            this();
            addGoal( goal );
        }

        @Override
        public String toString() {
            StringBuffer buf = new StringBuffer();
            buf.append( "goalStack: " );
            for ( FOS fos : goalStack ) {
                buf.append( fos );
                buf.append( " " );
            }
            buf.append( "\nsolution: " );
            buf.append( solution.toString() );
            buf.append( "\nlevel: " + level );
            return buf.toString();
        }

        public StackEntry copy( SubstitutionSet set ) {
            StackEntry entry = new StackEntry();
            entry.goalStack = new ArrayList<FOS>( goalStack );
            entry.solution = ( (SubstitutionSet) solution.clone() ).merge( set ).idempotencyBuilder().mostGeneralUnifier();
            entry.level = level + 1;
            return entry;
        }

        public void addGoal( Object object ) {
            level++;
            if ( object instanceof FOS ) {
                goalStack.add( (FOS) object );
            }
            else if ( object instanceof TerAND ) {
                for ( FOS formula : (TerAND) object ) {
                    goalStack.add( formula );
                }
            }
            else {
                throw new RuntimeException( "[QueryEngine] Should not reach here - goal value: " + object );
            }
        }

        public FOS getNextGoal() {
            return goalStack.remove( 0 ).apply( solution );
        }

        /**
         * remove all variables that were not in the initial goal.
         * 
         * @param variableSet
         */
        public void update( List<FOS> variableSet ) {
            boolean include = false;
            int i = 0;
            while ( i < solution.size() ) {
                Substitution sub = solution.getBinding( i );
                include = false;

                int j = 0;
                while ( !include && ( j < variableSet.size() ) ) {
                    include = sub.getFirst().isSameAs( variableSet.get( j ) );
                    j++;
                }

                if ( include ) {
                    i++;
                }
                else {
                    solution.removeBinding( i );
                }
            }
        }
    }

    public static class TestQuerySource implements Queryable {
        private List<Object> beliefs = new ArrayList<Object>();

        public TestQuerySource() {
            try {
                beliefs.add( LogicFactory.create( "BELIEF(clientConnection(jmcf.core.IPushData,test1.ucd.ie,TokenClient_5@A,output,test2.ucd.ie,TokenServer_2@B,input,TokenClient_5@APCTokenServer_2@BPSTokenClient_5@A,TCP))" ) );
            }
            catch ( Exception e ) {

            }
        }

        public List<Object> query( FOS goal ) {
            return beliefs;
        }

        public boolean hasType( String type ) {
            return type.equals( "BELIEF" );
        }
    }

    public static void main( String[] args ) {
        System.out.println( "test program" );
        QueryEngine engine = new QueryEngine();
        engine.addSource( new TestQuerySource() );
        try {
            System.out.println( "test: " + engine.queryAll( LogicFactory.create( "BELIEF(clientConnection(jmcf.core.IPushData,test1.ucd.ie,TokenClient_5@A,output,test2.ucd.ie,TokenServer_2@B,input,?adapter,?protocol))" ) ) );
            System.out.println( "sets: " + engine.getSubstitutionSets() );
        }
        catch ( Exception e ) {
            System.out.println( "error running logic engine" );
        }
        /*
         * try { SubstitutionSet set = null; FOS fos1 = (FOS)
         * LogicFactory.create("sosaaService(\"af.service.sosaa\")"); FOS fos2 =
         * (FOS) LogicFactory.create("sosaaService(af.service.sosaa)");
         * System.out.println("fos1: " + fos1); System.out.println("fos2: " +
         * fos2); if (fos2.isVariable()) { set = new SubstitutionSet();
         * set.addBinding(new Substitution(fos2,fos1)); } else { set =
         * Unifier.unifyFormulae(fos1, fos2); }
         * 
         * System.out.println("set: " + set);
         * 
         * } catch (Exception e) {
         * System.out.println("error running logic engine"); }
         */
    }
}
