/**
 * Title:       Unifier.java
 * Copyright:   Copyright (c) 1996-2004 The Agent Factory Working Group. All rights reserved.
 * Licence:     This file is free software; you can redistribute it and/or modify
 *              it under the terms of the GNU Lesser General Public License as published by
 *              the Free Software Foundation; either version 2.1, or (at your option)
 *              any later version.
 *
 *              This file is distributed in the hope that it will be useful,
 *              but WITHOUT ANY WARRANTY; without even the implied warranty of
 *              MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *              GNU Lesser General Public License for more details.
 *
 *              You should have received a copy of the GNU Lesser General Public License
 *              along with Agent Factory; see the file COPYING.  If not, write to
 *              the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *              Boston, MA 02111-1307, USA.
 */

package com.agentfactory.logic.reasoner;

import com.agentfactory.logic.lang.FOS;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * 
 * @author Terry Lowen
 * @version July 2002
 */
public class Unifier {
    private static final Logger logger = Logger.getLogger( Unifier.class.getName() );

    static {
        logger.setLevel( Level.WARNING );
    }

    private int variableCounter;
    private List<Substitution> stack;
    private SubstitutionSet solution;

    public Unifier() {
        variableCounter = 1;
    }

    private void initializeSolution() {
        solution = new SubstitutionSet();
        stack = new ArrayList<Substitution>();
    }

    public String solutionAsString() {
        if ( solution.isEmpty() ) {
            return "SOLUTION IS EMPTY";
        }

        StringBuffer ans = new StringBuffer();
        for ( Substitution substitution : solution ) {
            ans.append( "\n" + substitution.toString() );
        }

        return ans.toString();
    }

    public void incrementVariableCounter() {
        variableCounter = variableCounter + 1;
    }

    private boolean setUp( FOS f1, FOS f2 ) {
        if ( f1.numArguments() != f2.numArguments() ) {
            return false;
        }

        for ( int i = 0; i < f1.numArguments(); i++ ) {
            stack.add( new Substitution( f1.argAt( i ), f2.argAt( i ) ) );
        }

        return true;
    }

    /**
     * Perform a unification step...
     * 
     * @return
     */
    private boolean unification() {
        boolean matches = true;
        logger.log( Level.INFO, "Starting Unification: " + stack );

        while ( !stack.isEmpty() && matches ) {
            Substitution binding = stack.remove( stack.size() - 1 ).apply( solution );
            logger.log( Level.INFO, "Checking: " + binding + "..." );

            if ( binding.getFirst().isDistinctConstant( binding.getSecond() ) ) {
                logger.log( Level.INFO, " Distinct Constants" );
                matches = false;
            }
            else if ( binding.getFirst().isDistinctFunction( binding.getSecond() ) ) {
                logger.log( Level.INFO, " Distinct Function" );
                matches = false;
            }
            else if ( binding.getFirst().isDifferentTo( binding.getSecond() ) ) {
                logger.log( Level.INFO, " Different?" );
                matches = false;
            }
            else if ( binding.getFirst().isCompatibleFunction( binding.getSecond() ) ) {
                logger.log( Level.INFO, " Compatible Function" );
                if ( !setUp( binding.getFirst(), binding.getSecond() ) ) {
                    logger.log( Level.INFO, "  Setup Failure" );
                    matches = false;
                }
            }
            else if ( binding.getFirst().isVariable() && binding.getSecond().isVariable() ) {
                logger.log( Level.INFO, " Both Variables" );
                solution.addBinding( binding );
            }
            else if ( binding.getFirst().isNotFunctional( binding.getSecond() ) ) {
                logger.log( Level.INFO, " Not Functional" );
                matches = false;
            }
            else if ( binding.getFirst().isVariable() ) {
                logger.log( Level.INFO, " First is variable" );
                solution.addBinding( binding );
            }
            else if ( binding.getSecond().isVariable() ) {
                logger.log( Level.INFO, " Second is variable" );
                solution.addBinding( binding.inverse() );
            }

            logger.log( Level.INFO, "End of Step" );
        }

        SubstitutionSet subs = solution.idempotencyBuilder();
        while ( !subs.equals( solution ) ) {
            solution = subs;
            subs = solution.idempotencyBuilder();
        }

        return matches;
    }

    /**
     * Generate a set of bindings for two first-order formulae
     * 
     * @param f1
     * @param f2
     * @return
     */
    public static SubstitutionSet unifyFormulae( FOS f1, FOS f2 ) {
        try {
            Unifier u = new Unifier();
            u.initializeSolution();
            if ( u.setUp( f1, f2 ) ) {
                if ( u.unification() ) {
                    return u.solution;
                }
            }
        }
        catch ( Throwable th ) {
            th.printStackTrace();
            System.out.println( "Failed to unify: " + f1.toString() );
            System.out.println( "             to: " + f2.toString() );
        }

        return null;
    }

    public static void main( String[] args ) {
        System.out.println( Unifier.unifyFormulae( new FOS( "BELIEF(clientConnection(jmcf.core.IPushData,test1.ucd.ie,TokenClient_5@A,output,test2.ucd.ie,TokenServer_2@B,input,TokenClient_5@APCTokenServer_2@BPSTokenClient_5@A,TCP))" ), new FOS( "BELIEF(clientConnection(jmcf.core.IPushData,test1.ucd.ie,TokenClient_5@A,output,test2.ucd.ie,TokenServer_2@B,input,?adapter,?protocol))" ) ) );
    }
}
