package com.agentfactory.logic.update.lang;

import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.agentfactory.logic.update.interpreter.CoreUtilities;
import com.agentfactory.logic.update.reasoner.Bindings;

/**
 *
 * @author Rem Collier
 */
public class Unifier {
    private static final String LOGNAME = Unifier.class.getName();
    
    static {
        Logger.getLogger(LOGNAME).setLevel(Level.WARNING);
    }
    
    private int variableCounter;
    private List<ITerm[]> stack;
    private Bindings solution;
    
    public Unifier() {
        variableCounter=1;
    }
    
    private void initializeSolution() {
        solution = new Bindings();
        stack = new LinkedList<ITerm[]>();
    }
    
    public void incrementVariableCounter() {
        variableCounter++;
    }
    
    private boolean setUp(Predicate p1, Predicate p2) {
    	if (!p1.functor().equals(p2.functor())) return false;
    	
        if (p1.size() != p2.size()) {
            return false;
        }
        
        for(int i=0; i < p1.size(); i++) {
            stack.add(new ITerm[] {p1.termAt(i),p2.termAt(i)});
        }
        
        return true;
    }
    
    private boolean setUp(Function f1, Function f2) {
        if (f1.size() != f2.size()) {
            return false;
        }
        
        for(int i=0; i < f1.size(); i++) {
            stack.add(new ITerm[] {f1.termAt(i),f2.termAt(i)});
        }
        
        return true;
    }
        
    private ITerm[] applySolution(ITerm[] terms, Bindings solution) {
    	terms[0].apply(solution);
    	terms[1].apply(solution);
    	return terms;
    }
    
    /**
     * Perform a unification step...
     * @return
     */
    private boolean unification() {
        boolean matches = true;
        
        while (!stack.isEmpty() && matches) {
        	ITerm[] term = stack.remove(stack.size()-1);
//        	System.out.println("\n\n_____________________________");
//        	System.out.println(CoreUtilities.presenter.toString(term[0]));
//        	System.out.println(CoreUtilities.presenter.toString(term[1]));
        	
            ITerm[] binding = applySolution(term, solution);
            
            if ( binding[0].isDistinctFunction( binding[1] ) ) {
                // Check if they have different functors
                matches = false;
            } else if ( binding[0].isCompatibleFunction( binding[1] ) ) {
                // Check if they have the same functor / num of terms 
                if ( !setUp( (Function) binding[0], (Function) binding[1] ) ) {
                    matches = false;
                }
            } else if (binding[0].isNotFunctional( binding[1] ) ) {
                matches = false;
            } else if ( Variable.class.isInstance(binding[0]) && Variable.class.isInstance(binding[1]) ) {
            	// Check if they are both variables
                solution.addBinding((Variable) binding[0], binding[1]);
            } else if ( Variable.class.isInstance(binding[0]) ) {
                solution.addBinding((Variable) binding[0], binding[1]);
            } else if ( Variable.class.isInstance(binding[1]) ) {
                solution.addBinding((Variable) binding[1], binding[0]);
            } else if (IntegerTerm.class.isInstance(binding[0]) && IntegerTerm.class.isInstance( binding[1] ) ) {
                matches = ((IntegerTerm) binding[0]).value() == ((IntegerTerm) binding[1]).value();
            } else if (RealTerm.class.isInstance(binding[0]) && RealTerm.class.isInstance( binding[1] ) ) {
                matches = ((RealTerm) binding[0]).value() == ((RealTerm) binding[1]).value();
            } else if (ListTerm.class.isInstance(binding[0]) && ListTerm.class.isInstance( binding[1] ) ) {
                if ( !setUp( (Function) binding[0], (Function) binding[1] ) ) {
                    matches = false;
                }
            } else {
//            	System.out.println("[Unifier] This case is not handled: " +
//            			CoreUtilities.presenter.toString(term[0]) + " " + 
//            			CoreUtilities.presenter.toString(term[1]));
            }
        }
        
        
//    	System.out.println("\n\n_____________________________BINDING GENERATION");
//    	System.out.println(CoreUtilities.presenter.toString(solution));
        Bindings subs = solution.idempotencyBuilder();
//    	System.out.println(">>> match: "+ CoreUtilities.presenter.toString(subs));
        while(!subs.equals(solution)) {
            solution = subs;
            subs = solution.idempotencyBuilder();
//        	System.out.println(">>> match: "+ CoreUtilities.presenter.toString(subs));
        }
//    	System.out.println("\n\n_____________________________END OF BINDING GENERATION");
        
        return matches;
    }
    

    /**
     * Generate a set of bindings for two first-order formulae
     * 
     * @param f1
     * @param f2
     * @return
     */
    public static Bindings unifyFormulae(Predicate p1, Predicate p2) {
        try {
            Unifier u = new Unifier();
            u.initializeSolution();
            if (u.setUp(p1,p2)) {
                if(u.unification()) {
                    return u.solution;
                }
            }
        } catch (Throwable th) {
            th.printStackTrace();
            System.out.println("Failed to unify: " + p1.toString());
            System.out.println("             to: " + p2.toString());
        }
        
        return null;
    }
}
