package com.agentfactory.platform.impl;

import com.agentfactory.platform.core.*;
import com.agentfactory.service.profiler.TraceEvent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Observable;

/**
 *
 * @author remcollier & daithi
 */
public abstract class AbstractAgent extends Observable implements IAgent {
    protected IAgentContainer container;
    protected String name;
    protected int state;
    private boolean endOfTimeSlice;
    protected long timeSlice;
    private AgentID agentID;
    private long iterations;
    private List<IMessage> inbox;
    private List<IMessage> outbox;
    private List<IMessage> messageQueue;
    private Map<String, IPlatformService> services;
    private List<IPlatformService> serviceList;
    
    public AbstractAgent(String name) {
        this.name = name;
        iterations = 0;
        state = IAgent.INITIATED;
        messageQueue = new LinkedList<IMessage>();
        inbox = new LinkedList<IMessage>();
        outbox = new LinkedList<IMessage>();
        services = new HashMap<String, IPlatformService>();
        serviceList = new LinkedList<IPlatformService>();
    }
    
    public String getName() {
        return name;
    }

    public int getState() {
        return state;
    }
    
    public void setState(int state) {
        this.state = state;
    }

    public void setContainer(IAgentContainer container) {
        this.container = container;
  
        // Bind to all available MTS:
        for (IPlatformService service: container.bindToMessageTransportServices(this)) {
            services.put(service.getName(), service);
            addToServiceList(service);
        }
        createAgentID();
    }
    
    private void addToServiceList(IPlatformService service) {
        int priority = service.getPriority();
        boolean added = false;
        int index = 0;
        while (!added && (index < serviceList.size())) {
            if (priority > serviceList.get(index).getPriority()) {
                serviceList.add(index, service);
                added = true;
            }
            index++;
        }
        
        if (!added) {
            serviceList.add(service);
        }
    }

    public IPlatformService getPlatformService(String id) {
        return services.get(id);
    }
    
    public Collection<IPlatformService> getBoundPlatformServices() {
        return serviceList;
    }
    
    @SuppressWarnings("unchecked")
    /**
     * Get a list of platform services of a given class
     * @param targetClass The class of services desired (must be a class that implements the IPlatformService interface)
     * @return A {@code List} of appropriate services found, ordered by priority
     */
    public <T extends IPlatformService> List<T> getPlatformServicesByClass(Class<T> targetClass) {
        List<T> list = new LinkedList<T>();
        for (IPlatformService service: serviceList) {
            if (targetClass.isInstance(service)) {
                list.add( (T) service);
            }
        }
        return list;
    }
    
    /**
     * Binds the agent to the specified platform service (if it exists) and
     * returns true (indicating success).  If the service does not exist, then
     * false is returned.
     * 
     * @param id the service id
     * @return
     */
    public boolean bindToPlatformService(String id) {
        if (services.containsKey(id)) {
            return true;
        }
        
        IPlatformService service = container.bindToPlatformService(this, id);
        if (service == null) {
            return false;
        }
        
        services.put(id, service);
        addToServiceList(service);
        return true;
    }

    public synchronized void setEndOfTimeSlice(boolean endOfTimeSlice) {
        this.endOfTimeSlice = endOfTimeSlice;
    }
    
    public synchronized boolean isEndOfTimeSlice() {
        return endOfTimeSlice;
    }
    
    /**
     * Set the current time slice for the agent
     * 
     * @param timeSlice
     */
    public void setTimeSlice(long timeSlice) {
        this.timeSlice = timeSlice;
    }
    
    /**
     * Return the current time slice for the agent
     * 
     * @return
     */
    public long getTimeSlice() {
        return timeSlice;
    }
    
    /**
     * Return the agents AID
     * @return
     */
    public AgentID getAgentID() {
        return agentID;
    }
    
    /**
     * Return the number of iterations of the interpreter cycle
     * @return
     */
    public long getIterations() {
        return iterations;
    }
    
    /**
     * Should be called when the agent has finished the current iteration of
     * its execution algorithm.
     */
    protected void endOfIteration() {
        iterations++;
        setChanged();
        notifyObservers(IAgent.END_OF_TIME_SLICE);
    }
    
    /**
     * This method is used by FIPA Message Transport Services to deliver
     * new messages received to the relevant agents.  The method has been
     * synchronized to ensure that no messages are added whilst the agent
     * is gathering beliefs about the currently received messages.
     *
     * @param message the message
     */
    public synchronized void receive(IMessage message) {
        messageQueue.add(message);
        TraceEvent evt = new TraceEvent("message/receive", this, message, System.currentTimeMillis(), 0);
        setChanged();
        notifyObservers(evt);
    }

    /**
     * This method copies any messages received into the inbox for processing by
     * the agent. Only messages received since the last invocation of this method
     * are copied into the inbox.
     *
     */
    protected synchronized void retrieveNewMessages() {
        inbox.clear();
        inbox.addAll(messageQueue);
        messageQueue.clear();
    }
    
    /**
     * Sends a message via one of the available MTS:
     * 1. For each receiver:
     *    a. For each address
     *       i.   Check the available MTS to see if the message can be sent via
     *            that address.
     *       ii.  if so, send the message and check on the success
     *       iii. if not, keep checking
     *    b. if the message was sent, stop checking addresses for this receiver
     * @param message the message to be sent
     * @return true if the message was sent, false otherwise
     */
    public boolean send(IMessage message) {
        outbox.add(message);
        
        TraceEvent evt = new TraceEvent("message/send", this, message, System.currentTimeMillis(), 0);
        setChanged();
        notifyObservers(evt);

        List<MessageTransportService> transportServices = getPlatformServicesByClass(MessageTransportService.class);
        for (AgentID aid: message.getReceivers()) {
            boolean sent = false;
            for (String address: aid.getAddresses()) {
                for (MessageTransportService mts: transportServices) {
                    if (mts.isValidService(address)) {
                        sent = mts.sendMessage(address, message);
                        break;
                    }
                }
                
                if (sent) {
                    break;
                }
            }
        }
        
        return true;
    }
    
    /**
     * Return a reference to the inbox
     * 
     * @return
     */
    public List<IMessage> getInbox() {
        return inbox;
    }
    
    /**
     * Return a reference to the outbox
     * 
     * @return
     */
    public List<IMessage> getOutbox() {
        return outbox;
    }
    
    /**
     * Create a new agent id for the agent
     */
    protected void createAgentID() {
        List<String> addresses = new LinkedList<String>();
        List<MessageTransportService> transportServices = getPlatformServicesByClass(MessageTransportService.class);
        for (MessageTransportService mts:transportServices) {
            addresses.add(mts.getAddress());
        }
        
        agentID = new AgentID(name, addresses, new ArrayList<String>());
    }
    
    public void terminate() {
    	try {
			container.remove(name);
		} catch (NoSuchAgentException e) {
			e.printStackTrace();
		}
    }
}
