package com.agentfactory.platform.impl;

import com.agentfactory.platform.core.IAgent;
import com.agentfactory.platform.core.IPlatformService;
import com.agentfactory.platform.core.IPlatformServiceManager;
import java.util.Collection;
import java.util.HashSet;
import java.util.Observable;
import java.util.Properties;
import java.util.Set;

/**
 * Abstract class to be extended by concrete implementations of Platform Services. The class is
 * declared abstract ONLY because it should not be implemented directly.
 * 
 * @author daithi
 */
public abstract class AbstractPlatformService extends Observable implements IPlatformService {
    protected IPlatformServiceManager manager;
    protected String name;
    protected Properties configuration;
    private int priority = IPlatformService.DEFAULT_PRIORITY;
    private Set<IAgent> boundAgents;
    
    public AbstractPlatformService() {
        boundAgents = new HashSet<IAgent>();
    }
    
    public void setPriority( int priority ) {
        this.priority = priority;
    }
    

    /**
     * This is an empty method that can be overridden in concrete implementations of
     * platform services.
     *
     * @param config The configuration to be used for thi service
     * @param manager The platform service manager
     */
    public void configure( Properties config, IPlatformServiceManager manager ) {
        this.configuration = config;
        this.manager = manager;
    }
    
    

    /**
     * This is an empty method that can be overridden in concrete implmentations of
     * platform services. It is called when the service is started.
     */
    protected void onStart() {
    }

    /**
     * This is an empty method that can be overridden in concrete implmentations of
     * platform services. It is called when the service is stopped.
     */
    protected void onStop() {
    }

    /**
     * This is an empty method that can be overridden in concrete implmentations of
     * platform services. It is called whenever an agent binds to the service using {@link #bind(com.agentfactory.platform.core.IAgent)}.
     */
    protected void onBind( IAgent agent ) {
    }

    /**
     * This is an empty method that can be overridden in concrete implmentations of
     * platform services. It is called when an agent unbinds from the service using {@link #unbind(com.agentfactory.platform.core.IAgent)}.
     */
    protected void onUnbind( IAgent agent ) {
    }

    /**
     * Get the priority of this service
     * @return The priority of the service
     */
    public int getPriority() {
        return priority;
    }

    /**
     * Set the name of this service
     * @param name The new name
     */
    public void setName( String name ) {
        this.name = name;
    }

    /**
     * Get the name of the service
     * @return The service's name
     */
    public String getName() {
        return name;
    }

    /**
     * Bind an agent to this service. This method may not be overridden. To perform additional tasks whenever 
     * an agent binds to the service, see the {@link AbstractPlatformService#onBind(com.agentfactory.platform.core.IAgent)} method.
     * @param agent The agent to be bound to this service.
     */
    public final void bind( IAgent agent ) {
        boundAgents.add( agent );
        onBind( agent );
        this.setChanged();
        this.notifyObservers();
    }

    /**
     * Bind an agent to this service. This method may not be overridden. To perform additional tasks whenever 
     * an agent binds to the service, see the {@link AbstractPlatformService#onUnbind(com.agentfactory.platform.core.IAgent)} method.
     * @param agent The agent to be unbound from this service.
     */
    public final void unbind( IAgent agent ) {
        boundAgents.remove( agent );
        onUnbind( agent );
        this.setChanged();
        this.notifyObservers();
    }

    public Collection<IAgent> getBoundAgents() {
        return boundAgents;
    }

    public IAgent getAgentByName( String name ) {
        IAgent toReturn = null;
        for ( IAgent a : boundAgents ) {
            if ( a.getName().equals( name ) ) {
                toReturn = a;
                break;
            }
        }
        return toReturn;
    }

    /**
     * Start the service. This method may not be overridden. To perform additional tasks whenever 
     * an the service is started, see the {@link #onStart()} method.
     */
    public final void start() {
        notifyObservers( "Service " + name + " started" );
        onStart();
    }

    /**
     * Stop the service. This method may not be overridden. To perform additional tasks whenever 
     * an the service is stopped, see the {@link #onStop()} method.
     */
    public final void stop() {
        onStop();
        this.notifyObservers( "Service " + name + " stopped" );
    }
    
    public IPlatformServiceManager getPlatformServiceManager() {
        return this.manager;
    }
    
    
}
