package com.agentfactory.rma;

/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

import com.agentfactory.platform.core.IMessage;
import com.agentfactory.platform.impl.AbstractAgent;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This class implements the core Reactive Message Agent architecture. This
 * simple reactive architecture provides a mechanism by which agents may
 * respond to incoming messages together with a simple memory that is based
 * on a HashMap.
 * 
 * Message handlers are stored in a Map that associates message performatives
 * with a List of MessageHandlers. When a message is processed, the performative
 * is used to select the appropriate set of handlers for handling the message.
 * These handlers are then searched, and the filter(...) method is used to
 * check whether or not the handler should be used to handle the message.  If
 * a handler is found, then the agent sends the message, otherwise, a
 * not understood message is used as the default response.
 * 
 * @author remcollier
 */
public abstract class ReactiveMessageAgent extends AbstractAgent {
    private Map<String, List<MessageHandler>> handlerMap;
    protected Map<String,Object> memory;
    private List<HandlerEvent> handledEvents;
    protected List<Event> eventQueue;
    protected List<Event> newEvents;
    private List<EventHandler> eventHandlers;

    /*
     * Constructor that creates a new Reactive Message Agent
     * 
     * @param name the name of the agent
     * @param platform the platform on which the agent has been created.
     */
    public ReactiveMessageAgent(String name) {
        super(name);
        
        Logger.getLogger(name).setLevel(Level.WARNING);
    
        handlerMap = new HashMap<String, List<MessageHandler>>();
        memory = new HashMap<String,Object>();
        
        handledEvents = new LinkedList<HandlerEvent>();
        newEvents = new LinkedList<Event>();

        eventQueue = new LinkedList<Event>();
        eventHandlers = new LinkedList<EventHandler>();
    }

    /**
     * The type of the agent, which for RMA agents is the string "RMA"
     * 
     * @return returns the string "RMA"
     */
    public String getType() {
        return "RMA";
    }

    /**
     * This method handles INITIALISE statements from the APS file.  The
     * expeced structure of this statement is:
     * 
     * INITIALISE <agt-name> <key>=<value>
     * 
     * Entries that match this format are automatically added to the memory
     * of the agent.
     * 
     * @param data the initialisation data
     */
    public void initialise(String data) {
        String[] prop = data.split("=");
        memory.put(prop[0].trim(), prop[1].trim());
    }

    /**
     * Invoked by the scheduler when the agent is scheduled to be executed.
     */
    public void execute() {
        IMessage message = null;
        Event event = null;
        
        // Step 1: Add any new events to the eventQueue
        eventQueue.addAll(newEvents);
        newEvents.clear();
        
        // Step 2: Keep processing messages until we run out of time.
        while ((!getInbox().isEmpty() || !eventQueue.isEmpty()) && !isEndOfTimeSlice()) {
            if (!getInbox().isEmpty()) {
                message = getInbox().remove(0);
                if (!handleMessage(message)) {
                    Logger.getLogger(name).log(Level.WARNING, "Message Not Understood: " + message);
                }
            }

            if (!eventQueue.isEmpty()) {
                event = eventQueue.remove(0);
                if (!handleEvent(event)) {
                    Logger.getLogger(name).log(Level.WARNING, "Unhandled Event: " + event);
                }
            }
        }
        
        endOfIteration();
    }
    
    /**
     * Return a reference to the agents memory
     * 
     * @return a map that is the agents memory
     */
    public Map<String, Object> getMemory() {
        return memory;
    }
    
    /**
     * Adds a new message handler to the handlerMap.  Performs lazy initialisation
     * of the associated list.
     * 
     * @param handler the handler to be added
     */
    public void addMessageHandler(MessageHandler handler) {
        String performative = handler.getPerformative().toUpperCase();
        List<MessageHandler> list = handlerMap.get(performative);
        if (list == null) {
            list = new LinkedList<MessageHandler>();
            handlerMap.put(performative, list);
        }
        
        list.add(handler);
    }
    
    public void addEventHandler(EventHandler handler) {
        eventHandlers.add(handler);
    }
    
    /**
     * Core message handling algorithm.  The list of relevant handlers is
     * retrieved from the handlerMap and then searched until either all
     * handlers are found to be irrelevent or a relevant handler is located.
     * 
     * In the event of a relevent handler being located, the handler is
     * then used to handle the message.  If this handler returns false, then
     * it signifies that the handler could not process the message, and hence
     * the message was not understood.
     * 
     * @param message the message to be handled
     * @return true if the message was understood, false otherwise
     */
    public boolean handleMessage(IMessage message) {
        boolean handled = false;
        
        List<MessageHandler> list = handlerMap.get(message.getPerformative().toUpperCase());
        if (list != null) {
            MessageHandler handler = null;
            Iterator it = list.iterator();
            while (!handled && it.hasNext()) {
                handler = (MessageHandler) it.next();
                handled = handler.filter(message);
            }
            
            if (handled) {
                long duration = System.currentTimeMillis();
                    
                handled = handler.act(message);
                    
                duration = System.currentTimeMillis() - duration;
                handledEvents.add(new HandlerEvent(handler.getClass().getName(), message.toString(), duration, handled));
            }
        }
        
        return handled;
    }
    
    public boolean handleEvent(Event event) {
        boolean handled = false;
        EventHandler handler = null;
        
        Iterator it = eventHandlers.iterator();
        while (!handled && it.hasNext()) {
            handler = (EventHandler) it.next();
            
            handled = handler.filter(event);
            if (handled) {
                long duration = System.currentTimeMillis();
                    
                handler.handle(event);
                    
                duration = System.currentTimeMillis() - duration;
                handledEvents.add(new HandlerEvent(handler.getClass().getName(), event.toString(), duration, handled));
            }
        }
        
        return handled;
    }

    /**
     * Returns a list of handled events that occurred since the last iteration
     * 
     * @return a list object
     */
    public List<HandlerEvent> getHandledEvents() {
        List<HandlerEvent> list = handledEvents;
        handledEvents = new LinkedList<HandlerEvent>();
        return list;
    }
    
    public List<String> getMessageHandlers() {
        List<String> list = new LinkedList<String>();
        for (List<MessageHandler> handlers:handlerMap.values()) {
            for (MessageHandler handler: handlers) {
                list.add(handler.getClass().getName());
            }
        }
        return list;
    }
    
    public List<String> getEventHandlers() {
        List<String> list = new LinkedList<String>();
        for (EventHandler handler: eventHandlers) {
            list.add(handler.getClass().getName());
        }
        return list;
    }
    
    public void addEvent(Event event) {
        newEvents.add(event);
    }
}
    
