package com.agentfactory.service.df;

import java.util.Calendar;
import java.util.HashSet;
import java.util.Set;

import com.agentfactory.platform.core.AgentID;

public class DFAgentDescription {
	public AgentID agentID;
	public Set<ServiceDescription> services;
	public Set<String> protocols;
	public Set<String> ontologies;
	public Set<String> languages;
	public Calendar leaseTime;
	public Set<String> scope;
	
	/**
	 * Create a default DFAgentDescription
	 * @param agentID
	 */
	public DFAgentDescription(AgentID agentID) {
		this(agentID, new HashSet<ServiceDescription>(), new HashSet<String>(), new HashSet<String>(), new HashSet<String>(), null, new HashSet<String>());
	}
	
	/**
	 * Create a new DFAgentDescription
	 * 
	 * @param agentID
	 * @param services
	 * @param protocols
	 * @param ontologies
	 * @param languages
	 * @param leaseTime
	 * @param scope
	 */
	public DFAgentDescription(AgentID agentID,
							Set<ServiceDescription> services, 
							Set<String> protocols, 
							Set<String> ontologies, 
							Set<String> languages, 
							Calendar leaseTime,
							Set<String> scope) {
		this.agentID =agentID;
		this.services = services;
		this.protocols = protocols;
		this.ontologies = ontologies;
		this.languages = languages;
		this.leaseTime = leaseTime;
		this.scope = scope;
	}
	
	/**
	 * Combine the existing DFAgentDescription with the one provided.
	 * 
	 * @param description
	 */
	public void merge(DFAgentDescription description) {
		for (ServiceDescription desc: description.services) {
			boolean found = false;
			for (ServiceDescription existing : services) {
				if (desc.name.equals(existing.name)) {
					existing.update(desc);
					found = true;
				}
			}
			if (!found) {
				services.add(desc);
			}
		}
	}

	/**
	 * Match the given DFAgentDescription with the one provided in the
	 * argument.
	 * 
	 * If any of the values in the argument are not matched to an equivalent
	 * value in this description, then fail the match.
	 * 
	 * @param description
	 * @return
	 */
	public boolean matches(DFAgentDescription description) {
		if (!agentID.matches(description.agentID)) return false;
		
		for (ServiceDescription dsd : description.services) {
			boolean match = false;
			for (ServiceDescription lsd : services) {
				if (lsd.matches(dsd)) {
					match = true;
				}
			}
			if (!match) return false;
		}

		if (!Utilities.matcher(protocols, description.protocols)) return false; 
		if (!Utilities.matcher(ontologies, description.ontologies)) return false; 
		if (!Utilities.matcher(languages, description.languages)) return false; 
//		Calendar leaseTime;
		if (!Utilities.matcher(scope, description.scope)) return false; 

		return true;
	}
	
	/**
	 * Add the given service to the set of services. if a service with this
	 * name already exists, then overwrite the existing service description.
	 * 
	 * @param description
	 */
	public void addService(ServiceDescription description) {
		boolean found = false;
		for (ServiceDescription existing : services) {
			if (description.name.equals(existing.name)) {
				existing.update(description);
				found = true;
			}
		}
		if (!found) {
			services.add(description);
		}
	}

	/**
	 * Remove the service identified by the given name
	 * 
	 * @param name
	 */
	public void removeService(String name) {
		ServiceDescription desc = null;
		for (ServiceDescription sd : services) {
			if (sd.name.equals(name)) {
				desc = sd;
			}
		}
		
		services.remove(desc);
	}

	/**
	 * Remove the given service description
	 * 
	 * @param desc
	 */
	public void removeService(ServiceDescription desc) {
		services.remove(desc);
	}
}
