package com.agentfactory.service.mts.http;

import com.agentfactory.platform.core.Constants;
import com.agentfactory.platform.core.IMessage;
import com.agentfactory.platform.core.TransportMessage;
import com.agentfactory.platform.impl.MessageTransportService;
import com.agentfactory.platform.impl.XMLEnvelope;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.ConnectException;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.Socket;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Calendar;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Implementation of the FIPA HTTP Message Transport Service.
 *
 * @author  Rem Collier
 */
public class HTTPMessageTransportService extends MessageTransportService {

    public static final String NAME = "fipa.mts.mtp.http.std";
    public static final String BOUNDARY = "boundary";
    private static final String LOGNAME = HTTPMessageTransportService.class.getName();

    static {
        Logger.getLogger( LOGNAME ).setLevel( Level.SEVERE );
    }
    private int port;
    private HTTPServer server;
    private Thread thread;

    /**
     * Checks if a given address is the local platform.  If it isn't, then
     * if forwards the message on to the local platform.
     *
     * @param address the address to be checked
     * @return true if the address corresponds to the local platform, false
     * otherwise.
     */
    private boolean isLocal( String address ) {
        URL url = null;
        try {
            url = new URL( address );
        }
        catch ( MalformedURLException mue ) {
            return false;
        }

        if ( url.getPort() != server.getPort() ) {
            return false;
        }

        String host = server.getHost();
        String hostname = null;
        try {
            hostname = InetAddress.getByName( server.getHost() ).getHostName();
        }
        catch ( UnknownHostException uhe ) {
            System.out.println( "unknown host" );
        }

        if ( ( url.getHost().equals( "localhost" ) ) ||
             ( url.getHost().equals( "127.0.0.1" ) ) ) {
            return true;
        }

        return ( ( url.getHost().equals( host ) ) ||
                 ( url.getHost().equals( hostname ) ) );
    }

    /**
     * Checks if a given address is valid for this service.
     *
     * @param address the address to be checked
     * @return true if the address is valid for the service. false otherwise.
     */
    public boolean isValidService( String address ) {
        URL url = null;
        try {
            url = new URL( address );
        }
        catch ( MalformedURLException mue ) {
            return false;
        }

        return ( url.getProtocol().equals( "http" ) );
    }

    /**
     * This method is responsible for transmitting messages using the FIPA
     * HTTP Message Transport Service.
     *
     * @param address a String object
     * @param message a Message object
     */
    public boolean sendMessage( String address, IMessage message ) {
        boolean sent = false;

        URL url = null;
        try {
            url = new URL( address );
        }
        catch ( MalformedURLException mue ) {
            return false;
        }

        if ( isLocal( address ) ) {
            Logger.getLogger( LOGNAME ).log( Level.INFO, "Local Receiver: " + message );

            sent = receiveMessage( message );
        }
        else {
            Logger.getLogger( LOGNAME ).log( Level.INFO, "Remote Receiver: " + message );
            TransportMessage transportMessage = createTransportMessage( message );

            int count = 0;
            while ( !sent && ( count < 5 ) ) {
                sent = sendMessage( url, transportMessage );
                count++;
            }
        }

        if ( !sent ) {
            Logger.getLogger( LOGNAME ).log( Level.WARNING, "Message Could not be Delivered: " + message );
        }
        else {
            Logger.getLogger( LOGNAME ).log( Level.INFO, "Message Delivered: " + message );
        }

        return sent;
    }

    public TransportMessage createTransportMessage( IMessage message ) {
        byte[] byteArray = message.toByteArray();
        XMLEnvelope envelope = new XMLEnvelope();
        envelope.setTo( message.getReceivers() );
        envelope.setIntendedReceiver( message.getReceivers() );
        envelope.setFrom( message.getSender() );
        envelope.setAclRepresentation( message.getAclRepresentation() );
        envelope.setDate( Calendar.getInstance() );
        envelope.setPayloadEncoding( message.getPayloadEncoding() );
        envelope.setPayloadLength( byteArray.length );
        return new TransportMessage( envelope, message );
    }

    /**
     * This method is used to send the message to all relevant destinations.
     * WARNING: Probably will have to change this interface when the client
     * is integrated with higher level code.
     *
     * @param url the destination url.
     * @param message the transport message.
     */
    private boolean sendMessage( URL url, TransportMessage message ) {
        boolean success = false;

        BufferedOutputStream outputStream = null;
        InputStreamReader reader = null;
        try {
            Socket s = new Socket( url.getHost(), url.getPort() );

            outputStream = new BufferedOutputStream( s.getOutputStream() );
            reader = new InputStreamReader( s.getInputStream() );

            PrintWriter out = new PrintWriter( outputStream, false );
            BufferedReader in = new BufferedReader( reader );

            // Sending header!
            out.print( "POST " + url.toString() + " HTTP/1.1" );
            out.print( Constants.CRLF );
            out.print( "Cache-Control: no-cache" );
            out.print( Constants.CRLF );
            out.print( "Host: " + url.getHost() + ":" + url.getPort() );
            out.print( Constants.CRLF );
            out.print( "MIME-Version: 1.0" );
            out.print( Constants.CRLF );
            out.print( "Content-Type: multipart/mixed;" );
            out.print( Constants.CRLF );
            out.print( " boundary=\"" + BOUNDARY + "\"" );
            out.print( Constants.CRLF );
            out.print( "Connection: close" );
            out.print( Constants.CRLF );
            out.print( Constants.CRLF );

            // This is the first boundary.  It marks the start of the envelope
            // part of the transport message.
            out.print( Constants.CRLF );
            out.print( "--" + BOUNDARY );
            out.print( Constants.CRLF );
            out.print( "Content-Type: application/" + Constants.FIPA_XML_ENVELOPE_TYPE );
            out.print( Constants.CRLF );
            out.print( Constants.CRLF );
            out.print( message.getEnvelope().toString() );
            out.print( Constants.CRLF );


            // This is the second boundary.  It marks the end of the envelope,
            // and the start of the payload
            out.print( "--" + BOUNDARY );
            out.print( Constants.CRLF );
            out.print( Constants.CRLF );
            out.print( Constants.CONTENT_TYPE + ": application/" );
            out.print( message.getEnvelope().getAclRepresentation() );
            out.print( Constants.CRLF );
            out.print( Constants.CRLF );


            byte[] payload = message.getPayload();
            for ( int i = 0; i < payload.length; i++ ) {
                out.write( (int) payload[i] );
            }


            out.print( Constants.CRLF );
            out.print( "--" + BOUNDARY + "--" );
            out.print( Constants.CRLF );

            out.flush();

            StringTokenizer tok = null;
            String line = null, command = null, code = null;
            while ( ( line = in.readLine() ) != null ) {
                if ( line.trim().length() == 0 ) {
                    break;
                }
                tok = new StringTokenizer( line );
                command = tok.nextToken();

                if ( command.equals( Constants.HTTP11 ) ) {
                    code = tok.nextToken();
                }
            }

            if ( "200".equals( code ) ) {
                success = true;
            }

            in.close();
            out.close();
            outputStream.close();
            reader.close();
            s.close();
        }
        catch ( ConnectException ce ) {
            System.out.println( "Failed to connect to: " + url + " for: " + message.getEnvelope().getIntendedReceiver() );
        }
        catch ( UnknownHostException uhe ) {
            System.out.println( "Unknown host: " + url.getHost() + " for: " + message.getEnvelope().getIntendedReceiver() );
        }
        catch ( IOException ioe ) {
            System.out.println( "IO Exception connecting to: " + url + " for: " + message.getEnvelope().getIntendedReceiver() );
        }

        return success;
    }

    /**
     * This method creates a socket on the assigned port, and listens for
     * connection attempts.
     *
     * TODO: We really should raise an exception if/when a service fails to start
     */
    @Override
    public void onStart() {
        try {
            port = Integer.parseInt( configuration.getProperty( "port" ) );
            server = new HTTPServer( port, this.manager, this );
            thread = new Thread( server, "HTTP-MTS" );
            thread.start();
            Logger.getLogger( LOGNAME ).log( Level.INFO, "Service started on port: " + port );
        }
        catch ( NumberFormatException e ) {
            System.err.println( "HTTP MTS Failed to Start: Port was not configured correctly" );
        }
    }

    /**
     * This method closes down the server socket, stopping any additional
     * connections being made to this message transport service.
     */
    @Override
    public void onStop() {
        server.setActive( false );
    }

    public String getAddress() {
        return "http://" + server.getHost() + ":" + server.getPort() + Constants.ACC_PATH;
    }

    public String getServiceID() {
        return server.getHost() + "-" + server.getPort();
    }
}
