package com.agentfactory.teleoreactive.interpreter.mentalstate;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.agentfactory.clf.interpreter.CoreUtilities;
import com.agentfactory.clf.lang.IFormula;
import com.agentfactory.clf.lang.IMPLIES;
import com.agentfactory.clf.lang.Predicate;
import com.agentfactory.clf.reasoner.Bindings;
import com.agentfactory.clf.reasoner.ForwardChainer;
import com.agentfactory.clf.reasoner.IQueryable;
import com.agentfactory.clf.reasoner.QueryResult;
import com.agentfactory.clf.reasoner.ResolutionBasedQueryEngine;
import com.agentfactory.platform.core.AgentID;

public class ModelTower implements IQueryable {
	/** The sensory percepts of the Agent. */
	private List<Predicate> percepts;
	/** The beliefs derived from the percepts of the Agent. */
	private List<Predicate> derivedBeliefs;
	/** All sensory and derived beliefs in map to allow faster access */
	private Map<String, List<Predicate>> beliefMap;
	/** A forward chaining resolution engine */
	private ForwardChainer perceptionTower;
	/** A list of the results of all the rules added to the Perception Tower */
	private List<Predicate> ruleResults;
	/** A list of beliefs received from other agents **Persistent** */
	private List<Predicate> informedBeliefs;
	/** An IQueryable where contradiction rules are stored. */
	private RuleBase theTruth;
	/** The result portions of the contradiction rules, used for checking for contradictions. */
	private List<Predicate> contradictionResults;
	/** is true when something has been changed in the informed beliefs, returns to false when a new model is derived */
	private boolean informedBeliefChange;
	/** Beliefs pertaining to messaging such as who we are communicating with etc. */
	private List<Predicate> messageBeliefs;
	private List<Predicate> newBeliefs;

	public ModelTower() {
		percepts = new LinkedList<Predicate>();
		newBeliefs = new LinkedList<Predicate>();
		derivedBeliefs = new LinkedList<Predicate>();
		beliefMap = new HashMap<String, List<Predicate>>();
		perceptionTower = new ForwardChainer();
		perceptionTower.addSource(this);
		ruleResults = new LinkedList<Predicate>();
		informedBeliefs = new LinkedList<Predicate>();
		theTruth = new RuleBase();
		contradictionResults = new LinkedList<Predicate>();
		messageBeliefs = new LinkedList<Predicate>();
	}

	public void addBelief(Predicate belief) {
		newBeliefs.add(belief);
	}

	public void endOfPerception() {
		addBeliefs(newBeliefs);
		newBeliefs.clear();
	}

	/**
	 * This adds a List of Predicates, which encapsulate the current sensory input. If no sensory input has changed the function exits, however if there has
	 * been a change to the inputs, then all percepts and derived beliefs are cleared and and regenerated.
	 * 
	 * @param newPercepts
	 *            A list of all the agents current sensory percepts.
	 */
	private void addBeliefs(List<Predicate> newPercepts) {
		if (areBeliefsAltered(newPercepts)) {
			clearTower();
			for (Predicate newBelief : newPercepts) {
				addToBeliefMap(newBelief);
				percepts.add(newBelief);
			}
			for (Predicate p : informedBeliefs) {
				addToBeliefMap(p);
			}
			deriveNewBeliefs();
			for (Predicate p : derivedBeliefs) {
				addToBeliefMap(p);
			}
			for (Predicate p : messageBeliefs) {
				addToBeliefMap(p);
			}
			removeContradictions();
			// if(informedBeliefChange){
			// addBeliefs(newPercepts); //TODO see is this really necessary!!!
			// }
		}
	}

	/**
	 * Adds a persistent belief to the belief base, this belief should have come from another agent.
	 * 
	 * @param newInformedBelief
	 *            the belief to be stored.
	 * */
	public void addNewInformedBelief(Predicate newInformedBelief) {
		if (!informedBeliefs.contains(newInformedBelief)) {
			informedBeliefs.add(newInformedBelief);
			addToBeliefMap(newInformedBelief);
			informedBeliefChange = true;
			removeContradictions();
		}
	}

	public void addMesageBelief(Predicate newMessagBelief) {
		if (!messageBeliefs.contains(newMessagBelief))
			messageBeliefs.add(newMessagBelief);
		addToBeliefMap(newMessagBelief);
	}

	/**
	 * This method takes an IMPLIES rule which is used to remove contradictions.
	 * 
	 * @param rule
	 *            a rule of the form "belief(?x) & contradictoryBelief(?x) => remove(contradictoryBelief(?x))"
	 */
	public void addContradictionRule(IMPLIES rule) {
		theTruth.addRule(rule);
		contradictionResults.add((Predicate) rule.result());
	}

	/**
	 * Adds the rules used in the Perception Tower for forward chaining to occur. It also stored the results of the rules for querying.
	 * 
	 * @param rule
	 *            A rule of type IMPLIES
	 */
	public void addRule(IMPLIES rule) {
		perceptionTower.addRule(rule);
		if (!ruleResults.contains((Predicate) rule.result()))
			ruleResults.add((Predicate) rule.result());
	}

	/**
	 * This method determines weather or not there has been a change in the sensory percepts. This is used to determine weather or not we need to derive new
	 * beliefs.
	 * 
	 * @param newBeliefs
	 *            A list of the most recent sensory percepts from all sensors.
	 * @return true if the percepts are different from the last set of percepts received, and false if they are the same.
	 */
	private boolean areBeliefsAltered(List<Predicate> newBeliefs) {
		if (newBeliefs.size() != percepts.size())
			return true;
		else {
			if (percepts.containsAll(newBeliefs)) {
				return false;
			}
			return true;
		}
	}

	/**
	 * Clears all percepts from the tower, both sensory and derived.
	 */
	private void clearTower() {
		percepts.clear();
		beliefMap.clear();
		derivedBeliefs.clear();
	}

	/**
	 * method for adding beliefs to the belief map which eliminates duplication.
	 * 
	 * @param newBelief
	 *            the belief to be added
	 */
	private void addToBeliefMap(Predicate newBelief) {
		List<Predicate> list = beliefMap.get(newBelief.functor());
		if (list == null) {
			list = new LinkedList<Predicate>();
			list.add(newBelief);
			beliefMap.put(newBelief.functor(), list);
		} else {
			for (Predicate bel : list) {
				if (bel.equals(newBelief)) {
					return;
				}
			}
			list.add(newBelief);
		}
	}

	/**
	 * Refreshes the perception tower and queries the results of the rules it holds, any derived beliefs returned are then added to the derived Beliefs.
	 * */
	private void deriveNewBeliefs() {
		perceptionTower.refresh();
		derivedBeliefs = perceptionTower.getDerivedBeliefs();
	}

	/**
	 * This method removes contradictions which may have arisen between the derived model and informed beliefs.
	 */
	private void removeContradictions() {
		ResolutionBasedQueryEngine e = new ResolutionBasedQueryEngine();
		e.addSource(theTruth);
		e.addSource(this);
		for (Predicate q : contradictionResults) {
			QueryResult r = e.query(q);
			if (r.found) {
				for (Bindings b : r.bindings) {
					IFormula res = q.apply(b);
					IFormula toBeRemoved = CoreUtilities.factory.createFormula(CoreUtilities.presenter.toString(((Predicate) res).termAt(0)));
					if (toBeRemoved.getVariables().size() == 0) {
						informedBeliefs.remove(toBeRemoved);
						messageBeliefs.remove(toBeRemoved);
						informedBeliefChange = true;
						beliefMap.get(((Predicate) toBeRemoved).functor()).remove(toBeRemoved);
					}
				}
			}
		}
	}

	/**
	 * Provides a List of string representations of the current Beliefs.
	 */
	public List<String> getBeliefs() {
		List<String> list = new LinkedList<String>();
		for (Predicate p : percepts) {
			list.add(CoreUtilities.presenter.toString(p));
		}
		for (Predicate p : derivedBeliefs) {
			list.add(CoreUtilities.presenter.toString(p));
		}
		for (Predicate p : informedBeliefs) {
			list.add(CoreUtilities.presenter.toString(p));
		}
		for (Predicate p : messageBeliefs) {
			list.add(CoreUtilities.presenter.toString(p));
		}
		/*
		 * for(List<Predicate> l:beliefMap.values()){ for(Predicate p :l) list.add(Utilities.presenter.toString(p)); }
		 */
		return list;
	}

	/**
	 * Checks the types of formulas which can be queried.
	 */
	public boolean hasType(String type) {
		return "raw-logic".equals(type);
	}

	public List<IFormula> query(IFormula query) {
		List<IFormula> list = new LinkedList<IFormula>();
		if (Predicate.class.isInstance(query)) {
			Predicate predicate = (Predicate) query;
			List<Predicate> bels = beliefMap.get(predicate.functor());
			if (bels != null) {
				for (Predicate fos : bels) {
					if (fos.matches(predicate)) {
						list.add(fos);
					}
				}
			}
		}
		return list;
	}

	public static void main(String[] args) {
		ModelTower mt = new ModelTower();
		ResolutionBasedQueryEngine q = new ResolutionBasedQueryEngine();
		q.addSource(mt);
		// list.add(;
		// list.add((Predicate) CoreUtilities.factory.createFormula("goto(1,1)"));
		// mt.addBeliefs(list);
		// mt.addNewInformedBelief((Predicate) CoreUtilities.factory.createFormula("request(Boss, newTask(Cleaner))"));
		mt.addNewInformedBelief((Predicate) CoreUtilities.factory.createFormula("goto(1,1)"));
		mt.addNewInformedBelief((Predicate) CoreUtilities.factory.createFormula("goto(2,2)"));
		mt.addNewInformedBelief((Predicate) CoreUtilities.factory.createFormula("goto(2,3)"));
		QueryResult t = q.queryAll(CoreUtilities.factory.createFormula("goto(?x,?y)"));
		IFormula i = CoreUtilities.factory.createFormula("goto(?x,?y)");
		System.out.println("Result Found " + t.found);
		for (Bindings b : t.bindings) {
			System.out.println(CoreUtilities.presenter.toString(b));

		}
		// mt.addContradictionRule((IMPLIES) i);
		// mt.addContradictionRule((IMPLIES) CoreUtilities.factory
		// .createFormula("dusty(?x,?y)&obstructed(?x,?y) => remove(dusty(?x,?y))"));
		System.out.println(mt.getBeliefs());
		// mt.removeContradictions();
		// System.out.println(mt.getBeliefs());
	}

	public List<IMPLIES> getRules() {
		return perceptionTower.getRules();
	}

	public List<IMPLIES> getTruthRules() {
		return theTruth.getRules();
	}

	public void addNewRequestBelief(AgentID sender, String content) {
		messageBeliefs.add((Predicate) CoreUtilities.factory.createFormula("request(" + sender + "," + content + ")"));
	}

	public void clearInformedBeliefs() {
		informedBeliefs.clear();
	}

	public void removeInformedBelief(Predicate createFormula) {
		for (Predicate p : informedBeliefs) {
			if (p.equals(createFormula)) {
				informedBeliefs.remove(p);
				break;
			}

		}

	}

}
